"""
Enterprise Entity Module.

Provides entity base classes, identity management, lifecycle tracking,
and entity comparison for DDD architectures.

Example:
    # Define an entity
    class Order(Entity[OrderId]):
        def __init__(self, order_id: OrderId, customer_id: str):
            super().__init__(order_id)
            self.customer_id = customer_id
            self.items = []
        
        def add_item(self, item: OrderItem):
            self.items.append(item)
            self.mark_modified()
    
    # Use entity
    order = Order(OrderId("123"), customer_id="cust-456")
    order.add_item(OrderItem("product-1", 2))
"""

from __future__ import annotations

import copy
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import total_ordering
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

ID = TypeVar('ID', bound='EntityId')
E = TypeVar('E', bound='Entity')


class EntityError(Exception):
    """Entity error."""
    pass


class InvalidIdError(EntityError):
    """Invalid identity error."""
    pass


class EntityStateError(EntityError):
    """Entity state error."""
    pass


class EntityLifecycle(str, Enum):
    """Entity lifecycle state."""
    NEW = "new"
    CLEAN = "clean"
    MODIFIED = "modified"
    DELETED = "deleted"


@total_ordering
class EntityId(ABC, Hashable):
    """
    Base class for entity identity.
    """
    
    @property
    @abstractmethod
    def value(self) -> Any:
        """Get the identity value."""
        pass
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EntityId):
            return False
        return self.value == other.value
    
    def __lt__(self, other: object) -> bool:
        if not isinstance(other, EntityId):
            return NotImplemented
        return str(self.value) < str(other.value)
    
    def __hash__(self) -> int:
        return hash(self.value)
    
    def __str__(self) -> str:
        return str(self.value)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.value!r})"


@dataclass
class StringId(EntityId):
    """String-based identity."""
    _value: str
    
    @property
    def value(self) -> str:
        return self._value
    
    @classmethod
    def generate(cls) -> "StringId":
        """Generate a new ID."""
        return cls(str(uuid.uuid4()))
    
    @classmethod
    def from_string(cls, value: str) -> "StringId":
        """Create from string."""
        if not value:
            raise InvalidIdError("ID cannot be empty")
        return cls(value)


@dataclass
class UuidId(EntityId):
    """UUID-based identity."""
    _value: uuid.UUID
    
    def __init__(self, value: Optional[Union[str, uuid.UUID]] = None):
        if value is None:
            self._value = uuid.uuid4()
        elif isinstance(value, uuid.UUID):
            self._value = value
        else:
            try:
                self._value = uuid.UUID(value)
            except ValueError as e:
                raise InvalidIdError(f"Invalid UUID: {value}") from e
    
    @property
    def value(self) -> uuid.UUID:
        return self._value
    
    @classmethod
    def generate(cls) -> "UuidId":
        """Generate a new ID."""
        return cls(uuid.uuid4())


@dataclass
class IntId(EntityId):
    """Integer-based identity."""
    _value: int
    
    @property
    def value(self) -> int:
        return self._value
    
    @classmethod
    def from_int(cls, value: int) -> "IntId":
        """Create from integer."""
        if value <= 0:
            raise InvalidIdError("ID must be positive")
        return cls(value)


@dataclass
class CompositeId(EntityId):
    """Composite identity made of multiple parts."""
    parts: tuple
    
    @property
    def value(self) -> tuple:
        return self.parts
    
    def __init__(self, *parts: Any):
        if not parts:
            raise InvalidIdError("Composite ID must have at least one part")
        self.parts = parts
    
    def __str__(self) -> str:
        return ":".join(str(p) for p in self.parts)


@dataclass
class EntityMetadata:
    """Entity metadata."""
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    version: int = 1
    deleted_at: Optional[datetime] = None
    created_by: Optional[str] = None
    updated_by: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


class Entity(Generic[ID]):
    """
    Base entity class.
    
    Example:
        class User(Entity[UserId]):
            def __init__(self, user_id: UserId, name: str):
                super().__init__(user_id)
                self.name = name
    """
    
    def __init__(self, id: ID):
        self._id = id
        self._metadata = EntityMetadata()
        self._lifecycle = EntityLifecycle.NEW
        self._original_hash: Optional[int] = None
    
    @property
    def id(self) -> ID:
        """Get entity identity."""
        return self._id
    
    @property
    def metadata(self) -> EntityMetadata:
        """Get entity metadata."""
        return self._metadata
    
    @property
    def lifecycle(self) -> EntityLifecycle:
        """Get lifecycle state."""
        return self._lifecycle
    
    @property
    def is_new(self) -> bool:
        """Check if entity is new."""
        return self._lifecycle == EntityLifecycle.NEW
    
    @property
    def is_modified(self) -> bool:
        """Check if entity is modified."""
        return self._lifecycle == EntityLifecycle.MODIFIED
    
    @property
    def is_deleted(self) -> bool:
        """Check if entity is deleted."""
        return self._lifecycle == EntityLifecycle.DELETED
    
    @property
    def version(self) -> int:
        """Get entity version."""
        return self._metadata.version
    
    def mark_clean(self) -> None:
        """Mark entity as clean (persisted)."""
        self._lifecycle = EntityLifecycle.CLEAN
        self._original_hash = self._compute_hash()
    
    def mark_modified(self) -> None:
        """Mark entity as modified."""
        if self._lifecycle != EntityLifecycle.DELETED:
            self._lifecycle = EntityLifecycle.MODIFIED
            self._metadata.updated_at = datetime.now()
    
    def mark_deleted(self) -> None:
        """Mark entity as deleted."""
        self._lifecycle = EntityLifecycle.DELETED
        self._metadata.deleted_at = datetime.now()
    
    def increment_version(self) -> None:
        """Increment entity version."""
        self._metadata.version += 1
        self._metadata.updated_at = datetime.now()
    
    def has_changes(self) -> bool:
        """Check if entity has unsaved changes."""
        if self._lifecycle in (EntityLifecycle.NEW, EntityLifecycle.MODIFIED):
            return True
        
        if self._original_hash is None:
            return True
        
        return self._compute_hash() != self._original_hash
    
    def _compute_hash(self) -> int:
        """Compute hash of current state."""
        # Override in subclasses for proper comparison
        return hash(str(self.__dict__))
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Entity):
            return False
        return self._id == other._id
    
    def __hash__(self) -> int:
        return hash(self._id)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id})"


class AuditableEntity(Entity[ID]):
    """
    Entity with audit tracking.
    """
    
    def __init__(self, id: ID, created_by: Optional[str] = None):
        super().__init__(id)
        self._metadata.created_by = created_by
    
    def update_by(self, user_id: str) -> None:
        """Record update by user."""
        self._metadata.updated_by = user_id
        self.mark_modified()
    
    @property
    def created_by(self) -> Optional[str]:
        return self._metadata.created_by
    
    @property
    def updated_by(self) -> Optional[str]:
        return self._metadata.updated_by


class SoftDeleteEntity(Entity[ID]):
    """
    Entity with soft delete support.
    """
    
    @property
    def is_soft_deleted(self) -> bool:
        return self._metadata.deleted_at is not None
    
    def soft_delete(self) -> None:
        """Soft delete the entity."""
        self._metadata.deleted_at = datetime.now()
        self.mark_modified()
    
    def restore(self) -> None:
        """Restore a soft-deleted entity."""
        self._metadata.deleted_at = None
        self.mark_modified()


class VersionedEntity(Entity[ID]):
    """
    Entity with optimistic locking support.
    """
    
    def check_version(self, expected_version: int) -> bool:
        """Check if version matches."""
        return self._metadata.version == expected_version
    
    def assert_version(self, expected_version: int) -> None:
        """Assert version matches or raise error."""
        if not self.check_version(expected_version):
            raise EntityStateError(
                f"Version mismatch: expected {expected_version}, "
                f"got {self._metadata.version}"
            )


class EntityFactory(Generic[E, ID]):
    """
    Factory for creating entities.
    """
    
    def __init__(
        self,
        entity_class: Type[E],
        id_generator: Callable[[], ID],
    ):
        self._entity_class = entity_class
        self._id_generator = id_generator
    
    def create(self, **kwargs: Any) -> E:
        """Create a new entity."""
        entity_id = self._id_generator()
        return self._entity_class(entity_id, **kwargs)
    
    def create_with_id(self, id: ID, **kwargs: Any) -> E:
        """Create entity with specific ID."""
        return self._entity_class(id, **kwargs)


class EntityBuilder(Generic[E]):
    """
    Builder for entities.
    """
    
    def __init__(self, entity_class: Type[E]):
        self._entity_class = entity_class
        self._attributes: Dict[str, Any] = {}
        self._id: Optional[Any] = None
    
    def with_id(self, id: Any) -> "EntityBuilder[E]":
        """Set entity ID."""
        self._id = id
        return self
    
    def with_attribute(
        self,
        name: str,
        value: Any,
    ) -> "EntityBuilder[E]":
        """Set an attribute."""
        self._attributes[name] = value
        return self
    
    def with_attributes(
        self,
        **kwargs: Any,
    ) -> "EntityBuilder[E]":
        """Set multiple attributes."""
        self._attributes.update(kwargs)
        return self
    
    def build(self) -> E:
        """Build the entity."""
        if self._id is None:
            raise EntityError("Entity ID is required")
        
        return self._entity_class(self._id, **self._attributes)


class EntityRegistry:
    """
    Registry for tracking entity instances.
    """
    
    def __init__(self):
        self._entities: Dict[Any, Entity] = {}
        self._dirty: set = set()
        self._removed: set = set()
    
    def register(self, entity: Entity) -> None:
        """Register an entity."""
        self._entities[entity.id] = entity
    
    def get(self, id: Any) -> Optional[Entity]:
        """Get entity by ID."""
        return self._entities.get(id)
    
    def contains(self, id: Any) -> bool:
        """Check if entity is registered."""
        return id in self._entities
    
    def mark_dirty(self, entity: Entity) -> None:
        """Mark entity as dirty."""
        self._dirty.add(entity.id)
    
    def mark_removed(self, entity: Entity) -> None:
        """Mark entity for removal."""
        self._removed.add(entity.id)
    
    def get_dirty(self) -> List[Entity]:
        """Get dirty entities."""
        return [
            self._entities[id] for id in self._dirty
            if id in self._entities
        ]
    
    def get_removed(self) -> List[Entity]:
        """Get removed entities."""
        return [
            self._entities[id] for id in self._removed
            if id in self._entities
        ]
    
    def clear(self) -> None:
        """Clear registry."""
        self._entities.clear()
        self._dirty.clear()
        self._removed.clear()


class IdentityMap(Generic[ID, E]):
    """
    Identity map pattern for caching entities.
    """
    
    def __init__(self):
        self._map: Dict[ID, E] = {}
    
    def get(self, id: ID) -> Optional[E]:
        """Get entity by ID."""
        return self._map.get(id)
    
    def add(self, entity: E) -> None:
        """Add entity to map."""
        self._map[entity.id] = entity
    
    def remove(self, id: ID) -> None:
        """Remove entity from map."""
        self._map.pop(id, None)
    
    def contains(self, id: ID) -> bool:
        """Check if entity exists."""
        return id in self._map
    
    def get_or_add(
        self,
        id: ID,
        factory: Callable[[], E],
    ) -> E:
        """Get existing or create new entity."""
        if id not in self._map:
            self._map[id] = factory()
        return self._map[id]
    
    def clear(self) -> None:
        """Clear the map."""
        self._map.clear()
    
    def __len__(self) -> int:
        return len(self._map)


# Decorators
def entity(id_type: Type[EntityId]) -> Callable[[Type], Type]:
    """
    Class decorator to mark as entity.
    
    Example:
        @entity(UuidId)
        class Product:
            name: str
            price: float
    """
    def decorator(cls: Type) -> Type:
        cls._entity_id_type = id_type
        return cls
    
    return decorator


def auditable(cls: Type[E]) -> Type[E]:
    """
    Class decorator to add audit capabilities.
    
    Example:
        @auditable
        class Order(Entity[OrderId]):
            ...
    """
    original_init = cls.__init__
    
    def new_init(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        self._audit_log = []
    
    def add_audit(self, action: str, details: Optional[Dict] = None):
        self._audit_log.append({
            "action": action,
            "timestamp": datetime.now(),
            "details": details or {},
        })
    
    cls.__init__ = new_init
    cls.add_audit = add_audit
    
    return cls


def versioned(cls: Type[E]) -> Type[E]:
    """
    Class decorator to add version tracking.
    
    Example:
        @versioned
        class Document(Entity[DocumentId]):
            ...
    """
    original_mark_modified = cls.mark_modified
    
    def new_mark_modified(self):
        original_mark_modified(self)
        self.increment_version()
    
    cls.mark_modified = new_mark_modified
    
    return cls


# Factory functions
def create_string_id(value: Optional[str] = None) -> StringId:
    """Create a string ID."""
    if value:
        return StringId.from_string(value)
    return StringId.generate()


def create_uuid_id(value: Optional[str] = None) -> UuidId:
    """Create a UUID ID."""
    if value:
        return UuidId(value)
    return UuidId.generate()


def create_int_id(value: int) -> IntId:
    """Create an integer ID."""
    return IntId.from_int(value)


def create_composite_id(*parts: Any) -> CompositeId:
    """Create a composite ID."""
    return CompositeId(*parts)


def create_entity_factory(
    entity_class: Type[E],
    id_class: Type[ID] = UuidId,
) -> EntityFactory[E, ID]:
    """Create an entity factory."""
    return EntityFactory(
        entity_class,
        id_class.generate,
    )


def create_identity_map() -> IdentityMap:
    """Create an identity map."""
    return IdentityMap()


def create_entity_registry() -> EntityRegistry:
    """Create an entity registry."""
    return EntityRegistry()


__all__ = [
    # Exceptions
    "EntityError",
    "InvalidIdError",
    "EntityStateError",
    # Enums
    "EntityLifecycle",
    # Identity classes
    "EntityId",
    "StringId",
    "UuidId",
    "IntId",
    "CompositeId",
    # Data classes
    "EntityMetadata",
    # Entity classes
    "Entity",
    "AuditableEntity",
    "SoftDeleteEntity",
    "VersionedEntity",
    # Utility classes
    "EntityFactory",
    "EntityBuilder",
    "EntityRegistry",
    "IdentityMap",
    # Decorators
    "entity",
    "auditable",
    "versioned",
    # Factory functions
    "create_string_id",
    "create_uuid_id",
    "create_int_id",
    "create_composite_id",
    "create_entity_factory",
    "create_identity_map",
    "create_entity_registry",
]



"""
Enterprise Notification Hub Module.

Multi-channel notifications (email, SMS, push, Slack, webhooks)
with templating, scheduling, and delivery tracking.

Example:
    # Create notification hub
    hub = create_notification_hub()
    
    # Register channels
    hub.register_channel("email", EmailChannel(smtp_config))
    hub.register_channel("sms", SMSChannel(twilio_config))
    
    # Send notification
    await hub.send(
        user_id="user123",
        channels=["email", "push"],
        template="welcome",
        data={"name": "John"},
    )
    
    # Schedule notification
    await hub.schedule(
        user_id="user123",
        template="reminder",
        send_at=datetime.now() + timedelta(hours=24),
    )
"""

from __future__ import annotations

import asyncio
import functools
import json
import logging
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class NotificationError(Exception):
    """Notification error."""
    pass


class ChannelError(NotificationError):
    """Channel error."""
    pass


class DeliveryError(NotificationError):
    """Delivery error."""
    pass


class TemplateError(NotificationError):
    """Template error."""
    pass


class NotificationStatus(str, Enum):
    """Notification status."""
    PENDING = "pending"
    QUEUED = "queued"
    SENDING = "sending"
    SENT = "sent"
    DELIVERED = "delivered"
    FAILED = "failed"
    BOUNCED = "bounced"
    OPENED = "opened"
    CLICKED = "clicked"


class Priority(str, Enum):
    """Notification priority."""
    LOW = "low"
    NORMAL = "normal"
    HIGH = "high"
    URGENT = "urgent"


class ChannelType(str, Enum):
    """Channel types."""
    EMAIL = "email"
    SMS = "sms"
    PUSH = "push"
    SLACK = "slack"
    WEBHOOK = "webhook"
    IN_APP = "in_app"
    TEAMS = "teams"
    DISCORD = "discord"


@dataclass
class NotificationTemplate:
    """Notification template."""
    id: str
    name: str = ""
    subject: Optional[str] = None
    body: str = ""
    html_body: Optional[str] = None
    channel_templates: Dict[ChannelType, str] = field(default_factory=dict)
    variables: List[str] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Recipient:
    """Notification recipient."""
    id: str
    email: Optional[str] = None
    phone: Optional[str] = None
    device_tokens: List[str] = field(default_factory=list)
    slack_user_id: Optional[str] = None
    webhook_url: Optional[str] = None
    preferences: Dict[str, Any] = field(default_factory=dict)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class NotificationPayload:
    """Notification payload."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    template_id: Optional[str] = None
    recipient: Optional[Recipient] = None
    channels: List[ChannelType] = field(default_factory=list)
    subject: str = ""
    body: str = ""
    html_body: Optional[str] = None
    data: Dict[str, Any] = field(default_factory=dict)
    priority: Priority = Priority.NORMAL
    scheduled_at: Optional[datetime] = None
    expires_at: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DeliveryResult:
    """Delivery result."""
    notification_id: str
    channel: ChannelType
    status: NotificationStatus = NotificationStatus.PENDING
    provider_id: Optional[str] = None
    error: Optional[str] = None
    delivered_at: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class NotificationRecord:
    """Notification record."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    recipient_id: str = ""
    template_id: Optional[str] = None
    payload: Optional[NotificationPayload] = None
    results: List[DeliveryResult] = field(default_factory=list)
    status: NotificationStatus = NotificationStatus.PENDING
    created_at: datetime = field(default_factory=datetime.utcnow)
    sent_at: Optional[datetime] = None
    delivered_at: Optional[datetime] = None


# Channel interface
class NotificationChannel(ABC):
    """Abstract notification channel."""
    
    @property
    @abstractmethod
    def channel_type(self) -> ChannelType:
        """Get channel type."""
        pass
    
    @abstractmethod
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Send notification."""
        pass
    
    async def validate_recipient(self, recipient: Recipient) -> bool:
        """Validate recipient can receive notifications."""
        return True


class EmailChannel(NotificationChannel):
    """Email notification channel."""
    
    def __init__(
        self,
        smtp_host: str = "localhost",
        smtp_port: int = 587,
        username: Optional[str] = None,
        password: Optional[str] = None,
        from_address: str = "noreply@example.com",
        from_name: str = "Notification Service",
    ):
        self._smtp_host = smtp_host
        self._smtp_port = smtp_port
        self._username = username
        self._password = password
        self._from_address = from_address
        self._from_name = from_name
    
    @property
    def channel_type(self) -> ChannelType:
        return ChannelType.EMAIL
    
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Send email notification."""
        if not recipient.email:
            return DeliveryResult(
                notification_id=payload.id,
                channel=ChannelType.EMAIL,
                status=NotificationStatus.FAILED,
                error="No email address",
            )
        
        # Mock sending - in production, use aiosmtplib
        logger.info(f"Sending email to {recipient.email}: {payload.subject}")
        
        return DeliveryResult(
            notification_id=payload.id,
            channel=ChannelType.EMAIL,
            status=NotificationStatus.SENT,
            provider_id=str(uuid.uuid4()),
            delivered_at=datetime.utcnow(),
        )
    
    async def validate_recipient(self, recipient: Recipient) -> bool:
        return bool(recipient.email)


class SMSChannel(NotificationChannel):
    """SMS notification channel."""
    
    def __init__(
        self,
        provider: str = "mock",
        api_key: Optional[str] = None,
        from_number: Optional[str] = None,
    ):
        self._provider = provider
        self._api_key = api_key
        self._from_number = from_number
    
    @property
    def channel_type(self) -> ChannelType:
        return ChannelType.SMS
    
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Send SMS notification."""
        if not recipient.phone:
            return DeliveryResult(
                notification_id=payload.id,
                channel=ChannelType.SMS,
                status=NotificationStatus.FAILED,
                error="No phone number",
            )
        
        # Mock sending
        logger.info(f"Sending SMS to {recipient.phone}: {payload.body[:50]}")
        
        return DeliveryResult(
            notification_id=payload.id,
            channel=ChannelType.SMS,
            status=NotificationStatus.SENT,
            provider_id=str(uuid.uuid4()),
            delivered_at=datetime.utcnow(),
        )
    
    async def validate_recipient(self, recipient: Recipient) -> bool:
        return bool(recipient.phone)


class PushChannel(NotificationChannel):
    """Push notification channel."""
    
    def __init__(
        self,
        provider: str = "mock",
        api_key: Optional[str] = None,
    ):
        self._provider = provider
        self._api_key = api_key
    
    @property
    def channel_type(self) -> ChannelType:
        return ChannelType.PUSH
    
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Send push notification."""
        if not recipient.device_tokens:
            return DeliveryResult(
                notification_id=payload.id,
                channel=ChannelType.PUSH,
                status=NotificationStatus.FAILED,
                error="No device tokens",
            )
        
        # Mock sending
        logger.info(f"Sending push to {len(recipient.device_tokens)} devices")
        
        return DeliveryResult(
            notification_id=payload.id,
            channel=ChannelType.PUSH,
            status=NotificationStatus.SENT,
            provider_id=str(uuid.uuid4()),
            delivered_at=datetime.utcnow(),
            metadata={"devices": len(recipient.device_tokens)},
        )
    
    async def validate_recipient(self, recipient: Recipient) -> bool:
        return bool(recipient.device_tokens)


class SlackChannel(NotificationChannel):
    """Slack notification channel."""
    
    def __init__(
        self,
        bot_token: Optional[str] = None,
        webhook_url: Optional[str] = None,
    ):
        self._bot_token = bot_token
        self._webhook_url = webhook_url
    
    @property
    def channel_type(self) -> ChannelType:
        return ChannelType.SLACK
    
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Send Slack notification."""
        # Mock sending
        logger.info(f"Sending Slack message: {payload.body[:50]}")
        
        return DeliveryResult(
            notification_id=payload.id,
            channel=ChannelType.SLACK,
            status=NotificationStatus.SENT,
            provider_id=str(uuid.uuid4()),
            delivered_at=datetime.utcnow(),
        )


class WebhookChannel(NotificationChannel):
    """Webhook notification channel."""
    
    def __init__(
        self,
        default_url: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self._default_url = default_url
        self._headers = headers or {}
    
    @property
    def channel_type(self) -> ChannelType:
        return ChannelType.WEBHOOK
    
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Send webhook notification."""
        url = recipient.webhook_url or self._default_url
        
        if not url:
            return DeliveryResult(
                notification_id=payload.id,
                channel=ChannelType.WEBHOOK,
                status=NotificationStatus.FAILED,
                error="No webhook URL",
            )
        
        # Mock sending
        logger.info(f"Sending webhook to {url}")
        
        return DeliveryResult(
            notification_id=payload.id,
            channel=ChannelType.WEBHOOK,
            status=NotificationStatus.SENT,
            provider_id=str(uuid.uuid4()),
            delivered_at=datetime.utcnow(),
        )


class InAppChannel(NotificationChannel):
    """In-app notification channel."""
    
    def __init__(self):
        self._notifications: Dict[str, List[Dict]] = defaultdict(list)
    
    @property
    def channel_type(self) -> ChannelType:
        return ChannelType.IN_APP
    
    async def send(
        self,
        recipient: Recipient,
        payload: NotificationPayload,
    ) -> DeliveryResult:
        """Store in-app notification."""
        notification = {
            "id": payload.id,
            "subject": payload.subject,
            "body": payload.body,
            "data": payload.data,
            "created_at": datetime.utcnow().isoformat(),
            "read": False,
        }
        
        self._notifications[recipient.id].append(notification)
        
        return DeliveryResult(
            notification_id=payload.id,
            channel=ChannelType.IN_APP,
            status=NotificationStatus.DELIVERED,
            delivered_at=datetime.utcnow(),
        )
    
    async def get_notifications(
        self,
        user_id: str,
        unread_only: bool = False,
    ) -> List[Dict]:
        """Get user's in-app notifications."""
        notifications = self._notifications.get(user_id, [])
        
        if unread_only:
            return [n for n in notifications if not n["read"]]
        
        return notifications
    
    async def mark_read(self, user_id: str, notification_id: str) -> bool:
        """Mark notification as read."""
        for notification in self._notifications.get(user_id, []):
            if notification["id"] == notification_id:
                notification["read"] = True
                return True
        return False


# Template engine
class TemplateEngine:
    """Simple template engine."""
    
    def render(
        self,
        template: str,
        data: Dict[str, Any],
    ) -> str:
        """Render template with data."""
        result = template
        
        for key, value in data.items():
            result = result.replace(f"{{{{{key}}}}}", str(value))
            result = result.replace(f"{{{{ {key} }}}}", str(value))
        
        return result


class NotificationHub:
    """
    Notification hub service.
    """
    
    def __init__(
        self,
        template_engine: Optional[TemplateEngine] = None,
    ):
        self._channels: Dict[ChannelType, NotificationChannel] = {}
        self._templates: Dict[str, NotificationTemplate] = {}
        self._template_engine = template_engine or TemplateEngine()
        self._records: Dict[str, NotificationRecord] = {}
        self._scheduled: List[NotificationPayload] = []
        self._user_preferences: Dict[str, Dict[str, Any]] = {}
        self._hooks: Dict[str, List[Callable]] = defaultdict(list)
    
    def register_channel(
        self,
        channel: NotificationChannel,
    ) -> None:
        """Register notification channel."""
        self._channels[channel.channel_type] = channel
    
    def get_channel(
        self,
        channel_type: ChannelType,
    ) -> Optional[NotificationChannel]:
        """Get channel by type."""
        return self._channels.get(channel_type)
    
    def register_template(
        self,
        template: NotificationTemplate,
    ) -> None:
        """Register notification template."""
        self._templates[template.id] = template
    
    def add_hook(
        self,
        event: str,
        handler: Callable,
    ) -> None:
        """Add event hook."""
        self._hooks[event].append(handler)
    
    async def _trigger_hooks(
        self,
        event: str,
        *args,
        **kwargs,
    ) -> None:
        """Trigger event hooks."""
        for handler in self._hooks[event]:
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(*args, **kwargs)
                else:
                    handler(*args, **kwargs)
            except Exception as e:
                logger.error(f"Hook error: {e}")
    
    def set_user_preferences(
        self,
        user_id: str,
        preferences: Dict[str, Any],
    ) -> None:
        """Set user notification preferences."""
        self._user_preferences[user_id] = preferences
    
    def get_user_preferences(
        self,
        user_id: str,
    ) -> Dict[str, Any]:
        """Get user notification preferences."""
        return self._user_preferences.get(user_id, {})
    
    async def send(
        self,
        recipient: Union[str, Recipient],
        template: Optional[str] = None,
        channels: Optional[List[ChannelType]] = None,
        subject: Optional[str] = None,
        body: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
        priority: Priority = Priority.NORMAL,
        **kwargs,
    ) -> NotificationRecord:
        """
        Send notification.
        
        Args:
            recipient: Recipient ID or object
            template: Template ID
            channels: Channels to use
            subject: Notification subject
            body: Notification body
            data: Template data
            priority: Priority level
            
        Returns:
            Notification record
        """
        # Resolve recipient
        if isinstance(recipient, str):
            recipient_obj = Recipient(id=recipient)
        else:
            recipient_obj = recipient
        
        # Create payload
        payload = NotificationPayload(
            template_id=template,
            recipient=recipient_obj,
            channels=channels or [ChannelType.EMAIL],
            subject=subject or "",
            body=body or "",
            data=data or {},
            priority=priority,
            **kwargs,
        )
        
        # Apply template
        if template and template in self._templates:
            tmpl = self._templates[template]
            if not payload.subject and tmpl.subject:
                payload.subject = self._template_engine.render(
                    tmpl.subject, payload.data
                )
            if not payload.body:
                payload.body = self._template_engine.render(
                    tmpl.body, payload.data
                )
            if not payload.html_body and tmpl.html_body:
                payload.html_body = self._template_engine.render(
                    tmpl.html_body, payload.data
                )
        
        # Check user preferences
        prefs = self.get_user_preferences(recipient_obj.id)
        if prefs.get("muted"):
            return NotificationRecord(
                recipient_id=recipient_obj.id,
                template_id=template,
                payload=payload,
                status=NotificationStatus.FAILED,
            )
        
        # Create record
        record = NotificationRecord(
            recipient_id=recipient_obj.id,
            template_id=template,
            payload=payload,
        )
        
        # Trigger pre-send hook
        await self._trigger_hooks("pre_send", record)
        
        # Send to channels
        for channel_type in payload.channels:
            if channel_type not in self._channels:
                record.results.append(DeliveryResult(
                    notification_id=payload.id,
                    channel=channel_type,
                    status=NotificationStatus.FAILED,
                    error="Channel not registered",
                ))
                continue
            
            channel = self._channels[channel_type]
            
            # Validate recipient
            if not await channel.validate_recipient(recipient_obj):
                record.results.append(DeliveryResult(
                    notification_id=payload.id,
                    channel=channel_type,
                    status=NotificationStatus.FAILED,
                    error="Invalid recipient for channel",
                ))
                continue
            
            try:
                result = await channel.send(recipient_obj, payload)
                record.results.append(result)
            except Exception as e:
                record.results.append(DeliveryResult(
                    notification_id=payload.id,
                    channel=channel_type,
                    status=NotificationStatus.FAILED,
                    error=str(e),
                ))
        
        # Update status
        if all(r.status == NotificationStatus.SENT for r in record.results):
            record.status = NotificationStatus.SENT
            record.sent_at = datetime.utcnow()
        elif any(r.status == NotificationStatus.SENT for r in record.results):
            record.status = NotificationStatus.SENT
            record.sent_at = datetime.utcnow()
        else:
            record.status = NotificationStatus.FAILED
        
        # Store record
        self._records[record.id] = record
        
        # Trigger post-send hook
        await self._trigger_hooks("post_send", record)
        
        return record
    
    async def send_bulk(
        self,
        recipients: List[Union[str, Recipient]],
        template: Optional[str] = None,
        channels: Optional[List[ChannelType]] = None,
        **kwargs,
    ) -> List[NotificationRecord]:
        """Send bulk notifications."""
        tasks = [
            self.send(
                recipient=r,
                template=template,
                channels=channels,
                **kwargs,
            )
            for r in recipients
        ]
        
        return await asyncio.gather(*tasks, return_exceptions=True)
    
    async def schedule(
        self,
        recipient: Union[str, Recipient],
        send_at: datetime,
        **kwargs,
    ) -> NotificationPayload:
        """
        Schedule notification.
        
        Args:
            recipient: Recipient ID or object
            send_at: Scheduled send time
            **kwargs: Notification parameters
            
        Returns:
            Scheduled payload
        """
        if isinstance(recipient, str):
            recipient_obj = Recipient(id=recipient)
        else:
            recipient_obj = recipient
        
        payload = NotificationPayload(
            recipient=recipient_obj,
            scheduled_at=send_at,
            **kwargs,
        )
        
        self._scheduled.append(payload)
        return payload
    
    async def get_scheduled(
        self,
        before: Optional[datetime] = None,
    ) -> List[NotificationPayload]:
        """Get scheduled notifications."""
        if before:
            return [
                p for p in self._scheduled
                if p.scheduled_at and p.scheduled_at <= before
            ]
        return self._scheduled
    
    async def process_scheduled(self) -> List[NotificationRecord]:
        """Process due scheduled notifications."""
        now = datetime.utcnow()
        due = [
            p for p in self._scheduled
            if p.scheduled_at and p.scheduled_at <= now
        ]
        
        records = []
        for payload in due:
            if payload.recipient:
                record = await self.send(
                    recipient=payload.recipient,
                    template=payload.template_id,
                    channels=payload.channels,
                    subject=payload.subject,
                    body=payload.body,
                    data=payload.data,
                )
                records.append(record)
            
            self._scheduled.remove(payload)
        
        return records
    
    async def get_record(
        self,
        notification_id: str,
    ) -> Optional[NotificationRecord]:
        """Get notification record."""
        return self._records.get(notification_id)
    
    async def get_user_history(
        self,
        user_id: str,
        limit: int = 50,
    ) -> List[NotificationRecord]:
        """Get user notification history."""
        records = [
            r for r in self._records.values()
            if r.recipient_id == user_id
        ]
        records.sort(key=lambda r: r.created_at, reverse=True)
        return records[:limit]


# Factory functions
def create_notification_hub(
    template_engine: Optional[TemplateEngine] = None,
) -> NotificationHub:
    """Create notification hub."""
    return NotificationHub(template_engine)


def create_email_channel(
    smtp_host: str = "localhost",
    **kwargs,
) -> EmailChannel:
    """Create email channel."""
    return EmailChannel(smtp_host=smtp_host, **kwargs)


def create_sms_channel(
    provider: str = "mock",
    **kwargs,
) -> SMSChannel:
    """Create SMS channel."""
    return SMSChannel(provider=provider, **kwargs)


def create_push_channel(
    provider: str = "mock",
    **kwargs,
) -> PushChannel:
    """Create push channel."""
    return PushChannel(provider=provider, **kwargs)


def create_notification_template(
    id: str,
    subject: str = "",
    body: str = "",
    **kwargs,
) -> NotificationTemplate:
    """Create notification template."""
    return NotificationTemplate(id=id, subject=subject, body=body, **kwargs)


def create_recipient(
    id: str,
    email: Optional[str] = None,
    phone: Optional[str] = None,
    **kwargs,
) -> Recipient:
    """Create recipient."""
    return Recipient(id=id, email=email, phone=phone, **kwargs)


__all__ = [
    # Exceptions
    "NotificationError",
    "ChannelError",
    "DeliveryError",
    "TemplateError",
    # Enums
    "NotificationStatus",
    "Priority",
    "ChannelType",
    # Data classes
    "NotificationTemplate",
    "Recipient",
    "NotificationPayload",
    "DeliveryResult",
    "NotificationRecord",
    # Channels
    "NotificationChannel",
    "EmailChannel",
    "SMSChannel",
    "PushChannel",
    "SlackChannel",
    "WebhookChannel",
    "InAppChannel",
    # Engine
    "TemplateEngine",
    # Hub
    "NotificationHub",
    # Factory functions
    "create_notification_hub",
    "create_email_channel",
    "create_sms_channel",
    "create_push_channel",
    "create_notification_template",
    "create_recipient",
]



"""
Enterprise Deployment Manager Module.

Continuous deployment, blue-green deployments,
rolling updates, and canary releases.

Example:
    # Create deployment manager
    deployer = create_deployment_manager()
    
    # Create deployment
    deployment = await deployer.create(
        name="api-v2",
        version="2.0.0",
        strategy=DeploymentStrategy.BLUE_GREEN,
    )
    
    # Deploy
    await deployer.deploy(deployment.id)
    
    # Rollback if needed
    await deployer.rollback(deployment.id)
"""

from __future__ import annotations

import asyncio
import hashlib
import logging
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class DeploymentError(Exception):
    """Deployment error."""
    pass


class DeploymentFailed(DeploymentError):
    """Deployment failed."""
    pass


class RollbackFailed(DeploymentError):
    """Rollback failed."""
    pass


class DeploymentStrategy(str, Enum):
    """Deployment strategy."""
    ROLLING = "rolling"
    BLUE_GREEN = "blue_green"
    CANARY = "canary"
    RECREATE = "recreate"
    A_B_TEST = "a_b_test"


class DeploymentStatus(str, Enum):
    """Deployment status."""
    PENDING = "pending"
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    FAILED = "failed"
    ROLLED_BACK = "rolled_back"
    CANCELLED = "cancelled"


class HealthCheckStatus(str, Enum):
    """Health check status."""
    HEALTHY = "healthy"
    UNHEALTHY = "unhealthy"
    UNKNOWN = "unknown"


@dataclass
class DeploymentTarget:
    """Deployment target."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    host: str = ""
    port: int = 0
    weight: int = 1
    health_endpoint: str = "/health"
    status: HealthCheckStatus = HealthCheckStatus.UNKNOWN
    version: str = ""
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DeploymentConfig:
    """Deployment configuration."""
    name: str = ""
    version: str = ""
    strategy: DeploymentStrategy = DeploymentStrategy.ROLLING
    replicas: int = 3
    max_surge: int = 1
    max_unavailable: int = 0
    canary_percentage: int = 10
    health_check_interval: int = 10
    health_check_timeout: int = 5
    rollback_on_failure: bool = True
    auto_promote: bool = False
    promotion_delay: int = 300  # seconds
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Deployment:
    """Deployment."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    version: str = ""
    previous_version: Optional[str] = None
    strategy: DeploymentStrategy = DeploymentStrategy.ROLLING
    status: DeploymentStatus = DeploymentStatus.PENDING
    config: DeploymentConfig = field(default_factory=DeploymentConfig)
    targets: List[DeploymentTarget] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    rolled_back_at: Optional[datetime] = None
    progress: int = 0
    error: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DeploymentEvent:
    """Deployment event."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    deployment_id: str = ""
    event_type: str = ""
    message: str = ""
    timestamp: datetime = field(default_factory=datetime.utcnow)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DeploymentStats:
    """Deployment statistics."""
    total_deployments: int = 0
    successful_deployments: int = 0
    failed_deployments: int = 0
    rollbacks: int = 0
    avg_duration: float = 0.0


# Deployment store
class DeploymentStore(ABC):
    """Deployment storage."""
    
    @abstractmethod
    async def save(self, deployment: Deployment) -> None:
        pass
    
    @abstractmethod
    async def get(self, deployment_id: str) -> Optional[Deployment]:
        pass
    
    @abstractmethod
    async def delete(self, deployment_id: str) -> bool:
        pass
    
    @abstractmethod
    async def list_all(self) -> List[Deployment]:
        pass
    
    @abstractmethod
    async def add_event(self, event: DeploymentEvent) -> None:
        pass
    
    @abstractmethod
    async def get_events(self, deployment_id: str) -> List[DeploymentEvent]:
        pass


class InMemoryDeploymentStore(DeploymentStore):
    """In-memory deployment store."""
    
    def __init__(self):
        self._deployments: Dict[str, Deployment] = {}
        self._events: Dict[str, List[DeploymentEvent]] = {}
    
    async def save(self, deployment: Deployment) -> None:
        self._deployments[deployment.id] = deployment
    
    async def get(self, deployment_id: str) -> Optional[Deployment]:
        return self._deployments.get(deployment_id)
    
    async def delete(self, deployment_id: str) -> bool:
        if deployment_id in self._deployments:
            del self._deployments[deployment_id]
            return True
        return False
    
    async def list_all(self) -> List[Deployment]:
        return list(self._deployments.values())
    
    async def add_event(self, event: DeploymentEvent) -> None:
        if event.deployment_id not in self._events:
            self._events[event.deployment_id] = []
        self._events[event.deployment_id].append(event)
    
    async def get_events(self, deployment_id: str) -> List[DeploymentEvent]:
        return self._events.get(deployment_id, [])


# Health checker
class HealthChecker(ABC):
    """Health checker for targets."""
    
    @abstractmethod
    async def check(self, target: DeploymentTarget) -> HealthCheckStatus:
        pass


class HTTPHealthChecker(HealthChecker):
    """HTTP health checker."""
    
    def __init__(self, timeout: int = 5):
        self.timeout = timeout
    
    async def check(self, target: DeploymentTarget) -> HealthCheckStatus:
        try:
            # Simulated HTTP health check
            await asyncio.sleep(0.01)
            return HealthCheckStatus.HEALTHY
        except Exception:
            return HealthCheckStatus.UNHEALTHY


# Deployment strategies
class DeploymentStrategyImpl(ABC):
    """Deployment strategy implementation."""
    
    @abstractmethod
    async def deploy(
        self,
        deployment: Deployment,
        on_progress: Optional[Callable[[int], None]] = None,
    ) -> bool:
        pass
    
    @abstractmethod
    async def rollback(self, deployment: Deployment) -> bool:
        pass


class RollingDeployment(DeploymentStrategyImpl):
    """Rolling deployment strategy."""
    
    def __init__(self, health_checker: HealthChecker):
        self.health_checker = health_checker
    
    async def deploy(
        self,
        deployment: Deployment,
        on_progress: Optional[Callable[[int], None]] = None,
    ) -> bool:
        targets = deployment.targets
        total = len(targets)
        
        if total == 0:
            return True
        
        for idx, target in enumerate(targets):
            # Update target
            target.version = deployment.version
            target.status = HealthCheckStatus.UNKNOWN
            
            # Wait for health
            for _ in range(10):
                status = await self.health_checker.check(target)
                if status == HealthCheckStatus.HEALTHY:
                    target.status = status
                    break
                await asyncio.sleep(1)
            else:
                target.status = HealthCheckStatus.UNHEALTHY
                return False
            
            # Update progress
            progress = int((idx + 1) / total * 100)
            if on_progress:
                on_progress(progress)
        
        return True
    
    async def rollback(self, deployment: Deployment) -> bool:
        if not deployment.previous_version:
            return False
        
        for target in deployment.targets:
            target.version = deployment.previous_version
            target.status = HealthCheckStatus.UNKNOWN
            
            # Check health
            status = await self.health_checker.check(target)
            target.status = status
        
        return True


class BlueGreenDeployment(DeploymentStrategyImpl):
    """Blue-green deployment strategy."""
    
    def __init__(self, health_checker: HealthChecker):
        self.health_checker = health_checker
    
    async def deploy(
        self,
        deployment: Deployment,
        on_progress: Optional[Callable[[int], None]] = None,
    ) -> bool:
        # Deploy to "green" environment
        if on_progress:
            on_progress(10)
        
        # Prepare all targets
        for target in deployment.targets:
            target.version = deployment.version
            target.status = HealthCheckStatus.UNKNOWN
        
        if on_progress:
            on_progress(30)
        
        # Health check all targets
        for target in deployment.targets:
            status = await self.health_checker.check(target)
            target.status = status
            
            if status != HealthCheckStatus.HEALTHY:
                return False
        
        if on_progress:
            on_progress(70)
        
        # Switch traffic
        if on_progress:
            on_progress(100)
        
        return True
    
    async def rollback(self, deployment: Deployment) -> bool:
        if not deployment.previous_version:
            return False
        
        # Switch back to "blue"
        for target in deployment.targets:
            target.version = deployment.previous_version
        
        return True


class CanaryDeployment(DeploymentStrategyImpl):
    """Canary deployment strategy."""
    
    def __init__(
        self,
        health_checker: HealthChecker,
        canary_percentage: int = 10,
    ):
        self.health_checker = health_checker
        self.canary_percentage = canary_percentage
    
    async def deploy(
        self,
        deployment: Deployment,
        on_progress: Optional[Callable[[int], None]] = None,
    ) -> bool:
        targets = deployment.targets
        total = len(targets)
        
        if total == 0:
            return True
        
        # Calculate canary count
        canary_count = max(1, int(total * self.canary_percentage / 100))
        
        # Deploy canary
        canary_targets = targets[:canary_count]
        
        for target in canary_targets:
            target.version = deployment.version
            target.status = HealthCheckStatus.UNKNOWN
            
            status = await self.health_checker.check(target)
            target.status = status
            
            if status != HealthCheckStatus.HEALTHY:
                return False
        
        if on_progress:
            on_progress(30)
        
        # Monitor canary (simulated)
        await asyncio.sleep(0.1)
        
        if on_progress:
            on_progress(50)
        
        # Deploy rest
        remaining = targets[canary_count:]
        
        for idx, target in enumerate(remaining):
            target.version = deployment.version
            
            status = await self.health_checker.check(target)
            target.status = status
            
            if status != HealthCheckStatus.HEALTHY:
                return False
            
            progress = 50 + int((idx + 1) / len(remaining) * 50) if remaining else 100
            if on_progress:
                on_progress(progress)
        
        return True
    
    async def rollback(self, deployment: Deployment) -> bool:
        if not deployment.previous_version:
            return False
        
        for target in deployment.targets:
            target.version = deployment.previous_version
        
        return True


# Deployment manager
class DeploymentManager:
    """Deployment manager."""
    
    def __init__(
        self,
        store: Optional[DeploymentStore] = None,
        health_checker: Optional[HealthChecker] = None,
    ):
        self._store = store or InMemoryDeploymentStore()
        self._health_checker = health_checker or HTTPHealthChecker()
        self._strategies: Dict[DeploymentStrategy, DeploymentStrategyImpl] = {
            DeploymentStrategy.ROLLING: RollingDeployment(self._health_checker),
            DeploymentStrategy.BLUE_GREEN: BlueGreenDeployment(self._health_checker),
            DeploymentStrategy.CANARY: CanaryDeployment(self._health_checker),
            DeploymentStrategy.RECREATE: RollingDeployment(self._health_checker),
        }
        self._stats = DeploymentStats()
        self._listeners: List[Callable] = []
    
    async def create(
        self,
        name: str,
        version: str,
        strategy: DeploymentStrategy = DeploymentStrategy.ROLLING,
        targets: Optional[List[DeploymentTarget]] = None,
        config: Optional[DeploymentConfig] = None,
        **kwargs,
    ) -> Deployment:
        """Create deployment."""
        if config is None:
            config = DeploymentConfig(
                name=name,
                version=version,
                strategy=strategy,
            )
        
        deployment = Deployment(
            name=name,
            version=version,
            strategy=strategy,
            config=config,
            targets=targets or [],
            **kwargs,
        )
        
        await self._store.save(deployment)
        
        await self._add_event(deployment.id, "created", f"Deployment {name} v{version} created")
        
        self._stats.total_deployments += 1
        
        logger.info(f"Deployment created: {name} v{version}")
        
        return deployment
    
    async def deploy(
        self,
        deployment_id: str,
        on_progress: Optional[Callable[[int], None]] = None,
    ) -> bool:
        """Execute deployment."""
        deployment = await self._store.get(deployment_id)
        
        if not deployment:
            raise DeploymentError(f"Deployment not found: {deployment_id}")
        
        if deployment.status == DeploymentStatus.IN_PROGRESS:
            raise DeploymentError("Deployment already in progress")
        
        deployment.status = DeploymentStatus.IN_PROGRESS
        deployment.started_at = datetime.utcnow()
        await self._store.save(deployment)
        
        await self._add_event(deployment_id, "started", "Deployment started")
        
        # Get strategy
        strategy = self._strategies.get(deployment.strategy)
        
        if not strategy:
            deployment.status = DeploymentStatus.FAILED
            deployment.error = f"Unknown strategy: {deployment.strategy}"
            await self._store.save(deployment)
            raise DeploymentError(deployment.error)
        
        try:
            def update_progress(progress: int):
                deployment.progress = progress
                if on_progress:
                    on_progress(progress)
            
            success = await strategy.deploy(deployment, update_progress)
            
            if success:
                deployment.status = DeploymentStatus.COMPLETED
                deployment.completed_at = datetime.utcnow()
                self._stats.successful_deployments += 1
                
                await self._add_event(deployment_id, "completed", "Deployment completed successfully")
                
                logger.info(f"Deployment completed: {deployment.name}")
            else:
                deployment.status = DeploymentStatus.FAILED
                deployment.error = "Deployment failed"
                self._stats.failed_deployments += 1
                
                await self._add_event(deployment_id, "failed", "Deployment failed")
                
                if deployment.config.rollback_on_failure:
                    await self.rollback(deployment_id)
                
                logger.error(f"Deployment failed: {deployment.name}")
            
            await self._store.save(deployment)
            
            return success
        except Exception as e:
            deployment.status = DeploymentStatus.FAILED
            deployment.error = str(e)
            self._stats.failed_deployments += 1
            
            await self._store.save(deployment)
            await self._add_event(deployment_id, "error", str(e))
            
            if deployment.config.rollback_on_failure:
                await self.rollback(deployment_id)
            
            raise DeploymentFailed(str(e))
    
    async def rollback(self, deployment_id: str) -> bool:
        """Rollback deployment."""
        deployment = await self._store.get(deployment_id)
        
        if not deployment:
            raise DeploymentError(f"Deployment not found: {deployment_id}")
        
        await self._add_event(deployment_id, "rollback_started", "Rollback initiated")
        
        strategy = self._strategies.get(deployment.strategy)
        
        if not strategy:
            raise RollbackFailed(f"Unknown strategy: {deployment.strategy}")
        
        try:
            success = await strategy.rollback(deployment)
            
            if success:
                deployment.status = DeploymentStatus.ROLLED_BACK
                deployment.rolled_back_at = datetime.utcnow()
                self._stats.rollbacks += 1
                
                await self._add_event(deployment_id, "rollback_completed", "Rollback completed")
                
                logger.info(f"Rollback completed: {deployment.name}")
            else:
                await self._add_event(deployment_id, "rollback_failed", "Rollback failed")
                
                logger.error(f"Rollback failed: {deployment.name}")
            
            await self._store.save(deployment)
            
            return success
        except Exception as e:
            await self._add_event(deployment_id, "rollback_error", str(e))
            raise RollbackFailed(str(e))
    
    async def cancel(self, deployment_id: str) -> bool:
        """Cancel deployment."""
        deployment = await self._store.get(deployment_id)
        
        if not deployment:
            return False
        
        if deployment.status not in [DeploymentStatus.PENDING, DeploymentStatus.IN_PROGRESS]:
            return False
        
        deployment.status = DeploymentStatus.CANCELLED
        await self._store.save(deployment)
        
        await self._add_event(deployment_id, "cancelled", "Deployment cancelled")
        
        logger.info(f"Deployment cancelled: {deployment.name}")
        
        return True
    
    async def get(self, deployment_id: str) -> Optional[Deployment]:
        """Get deployment."""
        return await self._store.get(deployment_id)
    
    async def list_deployments(self) -> List[Deployment]:
        """List all deployments."""
        return await self._store.list_all()
    
    async def get_events(self, deployment_id: str) -> List[DeploymentEvent]:
        """Get deployment events."""
        return await self._store.get_events(deployment_id)
    
    async def _add_event(
        self,
        deployment_id: str,
        event_type: str,
        message: str,
    ) -> None:
        """Add deployment event."""
        event = DeploymentEvent(
            deployment_id=deployment_id,
            event_type=event_type,
            message=message,
        )
        
        await self._store.add_event(event)
        
        # Notify listeners
        for listener in self._listeners:
            try:
                if asyncio.iscoroutinefunction(listener):
                    await listener(event)
                else:
                    listener(event)
            except Exception as e:
                logger.error(f"Listener error: {e}")
    
    def add_listener(self, listener: Callable) -> None:
        """Add event listener."""
        self._listeners.append(listener)
    
    def get_stats(self) -> DeploymentStats:
        """Get statistics."""
        return self._stats


# Factory functions
def create_deployment_manager() -> DeploymentManager:
    """Create deployment manager."""
    return DeploymentManager()


def create_deployment(
    name: str,
    version: str,
    strategy: DeploymentStrategy = DeploymentStrategy.ROLLING,
    **kwargs,
) -> Deployment:
    """Create deployment."""
    return Deployment(name=name, version=version, strategy=strategy, **kwargs)


def create_deployment_target(
    name: str,
    host: str,
    port: int,
    **kwargs,
) -> DeploymentTarget:
    """Create deployment target."""
    return DeploymentTarget(name=name, host=host, port=port, **kwargs)


def create_deployment_config(
    name: str,
    version: str,
    **kwargs,
) -> DeploymentConfig:
    """Create deployment config."""
    return DeploymentConfig(name=name, version=version, **kwargs)


__all__ = [
    # Exceptions
    "DeploymentError",
    "DeploymentFailed",
    "RollbackFailed",
    # Enums
    "DeploymentStrategy",
    "DeploymentStatus",
    "HealthCheckStatus",
    # Data classes
    "DeploymentTarget",
    "DeploymentConfig",
    "Deployment",
    "DeploymentEvent",
    "DeploymentStats",
    # Stores
    "DeploymentStore",
    "InMemoryDeploymentStore",
    # Health checkers
    "HealthChecker",
    "HTTPHealthChecker",
    # Strategies
    "DeploymentStrategyImpl",
    "RollingDeployment",
    "BlueGreenDeployment",
    "CanaryDeployment",
    # Manager
    "DeploymentManager",
    # Factory functions
    "create_deployment_manager",
    "create_deployment",
    "create_deployment_target",
    "create_deployment_config",
]



"""
Task management module.

Thread-safe task registration, execution, and lifecycle tracking.
"""

from __future__ import annotations

import logging
import threading
import uuid
from typing import Any, Callable

from .exceptions import TaskExecutionError  # noqa: F401 - exported for library users

logger = logging.getLogger(__name__)


class Task:
    """Represents a single executable task."""

    __slots__ = ("id", "name", "objective", "executor", "inputs", "status", "result", "version")

    def __init__(
        self,
        name: str,
        objective: str,
        executor: Callable[..., Any],
        inputs: dict[str, Any] | None = None,
    ):
        self.id = str(uuid.uuid4())
        self.name = name
        self.objective = objective
        self.executor = executor
        self.inputs = inputs or {}
        self.status = "pending"
        self.result: Any = None
        self.version = "1.0.0"

    def run(self) -> Any:
        self.status = "running"
        logger.info("[Task:%s] Running", self.name)
        try:
            self.result = self.executor(**self.inputs)
            self.status = "completed"
            logger.info("[Task:%s] Completed successfully", self.name)
        except (TypeError, ValueError, KeyError, AttributeError) as e:
            self.status = "failed"
            logger.error("Task '%s' failed: %s", self.name, e)
        except Exception:  # noqa: BLE001
            self.status = "failed"
            logger.exception("Unexpected error in task '%s'", self.name)
        return self.result


class TaskManager:
    """Thread-safe task registry and executor."""

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self.tasks: dict[str, Task] = {}

    def register_task(self, task: Task) -> None:
        with self._lock:
            self.tasks[task.id] = task
        logger.info("[TaskManager] Registered '%s' (id=%s)", task.name, task.id)

    def get_task(self, task_id: str) -> Task | None:
        with self._lock:
            return self.tasks.get(task_id)

    def list_tasks(self) -> list[Task]:
        with self._lock:
            return list(self.tasks.values())

    def remove_task(self, task_id: str) -> None:
        with self._lock:
            if task_id in self.tasks:
                del self.tasks[task_id]
                logger.info("[TaskManager] Removed task %s", task_id)

    def run_all(self) -> dict[str, Any]:
        with self._lock:
            snapshot = dict(self.tasks)
        return {tid: task.run() for tid, task in snapshot.items()}

    def execute_task(self, task_name_or_id: str) -> Any:
        """Execute a task by name or ID."""
        with self._lock:
            task = self.tasks.get(task_name_or_id)
            if not task:
                task = next((t for t in self.tasks.values() if t.name == task_name_or_id), None)
        if task:
            return task.run()
        logger.warning("[TaskManager] Task '%s' not found", task_name_or_id)
        return None



"""
Database and Data Tools.

Tools for database operations and vector search.
"""

from .sql_tools import (
    MySQLRAGSearchTool,
    PostgreSQLRAGSearchTool,
    NL2SQLTool,
)
from .snowflake_tools import SnowflakeSearchTool, SingleStoreSearchTool
from .vector_tools import (
    QdrantVectorSearchTool,
    WeaviateVectorSearchTool,
    MongoDBVectorSearchTool,
)

__all__ = [
    # SQL Tools
    'MySQLRAGSearchTool',
    'PostgreSQLRAGSearchTool',
    'NL2SQLTool',
    # Snowflake Tools
    'SnowflakeSearchTool',
    'SingleStoreSearchTool',
    # Vector Tools
    'QdrantVectorSearchTool',
    'WeaviateVectorSearchTool',
    'MongoDBVectorSearchTool',
]



"""
Semantic Guardrail for content understanding and validation.

Uses pattern matching and heuristics for semantic validation.
Can be extended with embedding-based similarity.
"""

import re
from typing import Dict, Any, List, Optional, Tuple


class SemanticGuardrail:
    """
    Semantic guardrail for content understanding and validation.
    
    Uses pattern matching and heuristics for semantic validation.
    Can be extended with embedding-based similarity.
    """
    
    def __init__(self, 
                 name: str,
                 allowed_topics: Optional[List[str]] = None,
                 blocked_topics: Optional[List[str]] = None,
                 required_topics: Optional[List[str]] = None,
                 similarity_threshold: float = 0.7):
        self.name = name
        self.allowed_topics = allowed_topics or []
        self.blocked_topics = blocked_topics or []
        self.required_topics = required_topics or []
        self.similarity_threshold = similarity_threshold
        
        # Build keyword patterns
        self._blocked_patterns = [
            re.compile(rf'\b{re.escape(topic)}\b', re.IGNORECASE)
            for topic in self.blocked_topics
        ]
        self._required_patterns = [
            re.compile(rf'\b{re.escape(topic)}\b', re.IGNORECASE)
            for topic in self.required_topics
        ]
    
    def validate(self, content: str) -> Tuple[bool, List[str]]:
        """
        Validate content semantically.
        
        Returns:
            Tuple of (is_valid, list of violation messages)
        """
        violations = []
        
        # Check blocked topics
        for i, pattern in enumerate(self._blocked_patterns):
            if pattern.search(content):
                violations.append(
                    f"Content contains blocked topic: '{self.blocked_topics[i]}'"
                )
        
        # Check required topics
        for i, pattern in enumerate(self._required_patterns):
            if not pattern.search(content):
                violations.append(
                    f"Content missing required topic: '{self.required_topics[i]}'"
                )
        
        return len(violations) == 0, violations
    
    def compute_topic_score(self, content: str, topic: str) -> float:
        """Compute relevance score for a topic in content."""
        content_lower = content.lower()
        topic_lower = topic.lower()
        
        # Direct match
        if topic_lower in content_lower:
            return 1.0
        
        # Word overlap
        topic_words = set(topic_lower.split())
        content_words = set(content_lower.split())
        
        if not topic_words:
            return 0.0
        
        overlap = len(topic_words & content_words)
        return overlap / len(topic_words)


__all__ = ['SemanticGuardrail']



"""
Enterprise Streaming - Token-by-token streaming support.

Provides streaming capabilities for real-time LLM output,
progressive display, and efficient memory usage.

Features:
- Token-by-token streaming
- Async generators
- Progress callbacks
- Stream transformation
- Multi-consumer support
"""

import asyncio
import json
import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


# =============================================================================
# Stream Events
# =============================================================================

class StreamEventType(Enum):
    """Types of stream events."""
    START = "start"
    TOKEN = "token"
    CHUNK = "chunk"
    PROGRESS = "progress"
    COMPLETE = "complete"
    ERROR = "error"
    METADATA = "metadata"


@dataclass
class StreamEvent:
    """An event in a stream."""
    type: StreamEventType
    data: Any = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    timestamp: datetime = field(default_factory=datetime.now)
    
    # Token-specific
    token: Optional[str] = None
    token_id: Optional[int] = None
    
    # Progress
    tokens_generated: int = 0
    total_tokens: Optional[int] = None
    
    # Timing
    latency_ms: Optional[float] = None


# =============================================================================
# Stream Source
# =============================================================================

class StreamSource(ABC, Generic[T]):
    """Abstract base for stream sources."""
    
    @abstractmethod
    async def stream(self) -> AsyncGenerator[T, None]:
        """Generate stream items."""
        pass


class LLMStreamSource(StreamSource[StreamEvent]):
    """
    Stream source for LLM responses.
    
    Usage:
        >>> source = LLMStreamSource(llm, prompt="Write a story")
        >>> async for event in source.stream():
        ...     if event.type == StreamEventType.TOKEN:
        ...         print(event.token, end="", flush=True)
    """
    
    def __init__(
        self,
        llm: Any,
        prompt: str,
        model: Optional[str] = None,
        **kwargs,
    ):
        self.llm = llm
        self.prompt = prompt
        self.model = model
        self.kwargs = kwargs
        
        self._start_time: Optional[float] = None
        self._token_count = 0
    
    async def stream(self) -> AsyncGenerator[StreamEvent, None]:
        """Stream LLM response tokens."""
        self._start_time = time.time()
        self._token_count = 0
        
        # Emit start event
        yield StreamEvent(
            type=StreamEventType.START,
            metadata={"model": self.model, "prompt_length": len(self.prompt)},
        )
        
        try:
            # Stream from LLM
            async for token in self._stream_llm():
                self._token_count += 1
                latency = (time.time() - self._start_time) * 1000
                
                yield StreamEvent(
                    type=StreamEventType.TOKEN,
                    token=token,
                    tokens_generated=self._token_count,
                    latency_ms=latency,
                )
            
            # Emit complete event
            total_time = (time.time() - self._start_time) * 1000
            yield StreamEvent(
                type=StreamEventType.COMPLETE,
                tokens_generated=self._token_count,
                latency_ms=total_time,
                metadata={"tokens_per_second": self._token_count / (total_time / 1000)},
            )
            
        except Exception as e:
            yield StreamEvent(
                type=StreamEventType.ERROR,
                data=str(e),
                metadata={"error_type": type(e).__name__},
            )
    
    async def _stream_llm(self) -> AsyncGenerator[str, None]:
        """Stream from the LLM (implementation depends on LLM type)."""
        # Try different LLM interfaces
        if hasattr(self.llm, "astream"):
            async for chunk in self.llm.astream(self.prompt, **self.kwargs):
                if hasattr(chunk, "content"):
                    yield chunk.content
                elif isinstance(chunk, str):
                    yield chunk
        elif hasattr(self.llm, "stream"):
            for chunk in self.llm.stream(self.prompt, **self.kwargs):
                if hasattr(chunk, "content"):
                    yield chunk.content
                elif isinstance(chunk, str):
                    yield chunk
        else:
            # Fallback: simulate streaming from full response
            response = await self.llm.ainvoke(self.prompt, **self.kwargs)
            content = response.content if hasattr(response, "content") else str(response)
            
            for char in content:
                yield char


# =============================================================================
# Stream Processor
# =============================================================================

class StreamProcessor(Generic[T]):
    """
    Process and transform streams.
    
    Usage:
        >>> processor = StreamProcessor(source)
        >>> processor.add_transform(TokenToWordTransform())
        >>> processor.add_callback(print_token)
        >>> 
        >>> async for event in processor.process():
        ...     handle(event)
    """
    
    def __init__(self, source: StreamSource[T]):
        self.source = source
        self._transforms: List[Callable[[T], T]] = []
        self._callbacks: List[Callable[[T], Awaitable[None]]] = []
        self._filters: List[Callable[[T], bool]] = []
    
    def add_transform(self, transform: Callable[[T], T]) -> "StreamProcessor[T]":
        """Add a transformation function."""
        self._transforms.append(transform)
        return self
    
    def add_callback(self, callback: Callable[[T], Awaitable[None]]) -> "StreamProcessor[T]":
        """Add a callback for each item."""
        self._callbacks.append(callback)
        return self
    
    def add_filter(self, filter_fn: Callable[[T], bool]) -> "StreamProcessor[T]":
        """Add a filter function."""
        self._filters.append(filter_fn)
        return self
    
    async def process(self) -> AsyncGenerator[T, None]:
        """Process the stream with transforms and callbacks."""
        async for item in self.source.stream():
            # Apply filters
            if any(not f(item) for f in self._filters):
                continue
            
            # Apply transforms
            for transform in self._transforms:
                item = transform(item)
            
            # Call callbacks
            for callback in self._callbacks:
                await callback(item)
            
            yield item


# =============================================================================
# Stream Accumulator
# =============================================================================

class StreamAccumulator:
    """
    Accumulates stream tokens into complete content.
    
    Usage:
        >>> accumulator = StreamAccumulator()
        >>> async for event in stream:
        ...     accumulator.add(event)
        ...     print(f"\\rProgress: {accumulator.token_count}", end="")
        >>> print(accumulator.content)
    """
    
    def __init__(self):
        self._tokens: List[str] = []
        self._events: List[StreamEvent] = []
        self._start_time: Optional[float] = None
        self._end_time: Optional[float] = None
        self._error: Optional[str] = None
    
    def add(self, event: StreamEvent):
        """Add an event to the accumulator."""
        self._events.append(event)
        
        if event.type == StreamEventType.START:
            self._start_time = time.time()
        elif event.type == StreamEventType.TOKEN:
            if event.token:
                self._tokens.append(event.token)
        elif event.type == StreamEventType.COMPLETE:
            self._end_time = time.time()
        elif event.type == StreamEventType.ERROR:
            self._error = event.data
    
    @property
    def content(self) -> str:
        """Get accumulated content."""
        return "".join(self._tokens)
    
    @property
    def token_count(self) -> int:
        """Get token count."""
        return len(self._tokens)
    
    @property
    def duration_ms(self) -> Optional[float]:
        """Get stream duration in milliseconds."""
        if self._start_time and self._end_time:
            return (self._end_time - self._start_time) * 1000
        return None
    
    @property
    def tokens_per_second(self) -> Optional[float]:
        """Get tokens per second."""
        duration = self.duration_ms
        if duration and duration > 0:
            return self.token_count / (duration / 1000)
        return None
    
    @property
    def is_complete(self) -> bool:
        """Check if stream is complete."""
        return any(e.type == StreamEventType.COMPLETE for e in self._events)
    
    @property
    def has_error(self) -> bool:
        """Check if stream had an error."""
        return self._error is not None
    
    @property
    def error(self) -> Optional[str]:
        """Get error message if any."""
        return self._error


# =============================================================================
# Stream Multiplexer
# =============================================================================

class StreamMultiplexer:
    """
    Allows multiple consumers for a single stream.
    
    Usage:
        >>> mux = StreamMultiplexer(stream_source)
        >>> 
        >>> # Consumer 1
        >>> async def consumer1():
        ...     async for event in mux.subscribe():
        ...         process1(event)
        >>> 
        >>> # Consumer 2
        >>> async def consumer2():
        ...     async for event in mux.subscribe():
        ...         process2(event)
        >>> 
        >>> await asyncio.gather(mux.start(), consumer1(), consumer2())
    """
    
    def __init__(self, source: StreamSource[StreamEvent]):
        self.source = source
        self._queues: List[asyncio.Queue] = []
        self._started = False
        self._completed = False
    
    def subscribe(self) -> AsyncGenerator[StreamEvent, None]:
        """Subscribe to the stream."""
        queue: asyncio.Queue = asyncio.Queue()
        self._queues.append(queue)
        
        async def generator():
            while True:
                event = await queue.get()
                if event is None:
                    break
                yield event
        
        return generator()
    
    async def start(self):
        """Start streaming to all subscribers."""
        if self._started:
            return
        
        self._started = True
        
        async for event in self.source.stream():
            for queue in self._queues:
                await queue.put(event)
        
        # Signal completion
        for queue in self._queues:
            await queue.put(None)
        
        self._completed = True


# =============================================================================
# Stream Formatters
# =============================================================================

class StreamFormatter(ABC):
    """Base class for stream formatters."""
    
    @abstractmethod
    def format(self, event: StreamEvent) -> str:
        """Format a stream event."""
        pass


class SSEFormatter(StreamFormatter):
    """Format events as Server-Sent Events."""
    
    def format(self, event: StreamEvent) -> str:
        """Format as SSE."""
        data = {
            "type": event.type.value,
            "token": event.token,
            "tokens_generated": event.tokens_generated,
        }
        
        if event.type == StreamEventType.ERROR:
            data["error"] = event.data
        
        return f"data: {json.dumps(data)}\n\n"


class JSONLFormatter(StreamFormatter):
    """Format events as JSON Lines."""
    
    def format(self, event: StreamEvent) -> str:
        """Format as JSONL."""
        data = {
            "type": event.type.value,
            "timestamp": event.timestamp.isoformat(),
        }
        
        if event.token:
            data["token"] = event.token
        if event.data:
            data["data"] = event.data
        
        return json.dumps(data) + "\n"


class MarkdownFormatter(StreamFormatter):
    """Format tokens for markdown display."""
    
    def __init__(self):
        self._in_code_block = False
    
    def format(self, event: StreamEvent) -> str:
        """Format for markdown."""
        if event.type != StreamEventType.TOKEN:
            return ""
        
        token = event.token or ""
        
        # Track code blocks
        if "```" in token:
            self._in_code_block = not self._in_code_block
        
        return token


# =============================================================================
# Streaming Helpers
# =============================================================================

async def stream_to_string(
    stream: AsyncGenerator[StreamEvent, None],
) -> str:
    """Collect stream into a string."""
    accumulator = StreamAccumulator()
    
    async for event in stream:
        accumulator.add(event)
    
    return accumulator.content


async def stream_with_progress(
    stream: AsyncGenerator[StreamEvent, None],
    on_progress: Callable[[int, Optional[int]], None],
) -> str:
    """Stream with progress callback."""
    accumulator = StreamAccumulator()
    
    async for event in stream:
        accumulator.add(event)
        
        if event.type == StreamEventType.TOKEN:
            on_progress(event.tokens_generated, event.total_tokens)
    
    return accumulator.content


async def stream_with_display(
    stream: AsyncGenerator[StreamEvent, None],
    print_fn: Callable[[str], None] = None,
) -> str:
    """Stream with real-time display."""
    if print_fn is None:
        print_fn = lambda s: print(s, end="", flush=True)
    
    accumulator = StreamAccumulator()
    
    async for event in stream:
        accumulator.add(event)
        
        if event.type == StreamEventType.TOKEN and event.token:
            print_fn(event.token)
    
    return accumulator.content


# =============================================================================
# Stream Decorators
# =============================================================================

def stream_response(fn: Callable) -> Callable:
    """
    Decorator to automatically stream function output.
    
    Usage:
        >>> @stream_response
        >>> async def generate(prompt: str) -> str:
        ...     return await llm.invoke(prompt)
    """
    import functools
    
    @functools.wraps(fn)
    async def wrapper(*args, **kwargs) -> AsyncGenerator[StreamEvent, None]:
        start_time = time.time()
        
        yield StreamEvent(type=StreamEventType.START)
        
        try:
            result = await fn(*args, **kwargs)
            
            # Simulate streaming for non-streaming functions
            if isinstance(result, str):
                for i, char in enumerate(result):
                    yield StreamEvent(
                        type=StreamEventType.TOKEN,
                        token=char,
                        tokens_generated=i + 1,
                    )
            
            total_time = (time.time() - start_time) * 1000
            yield StreamEvent(
                type=StreamEventType.COMPLETE,
                tokens_generated=len(result) if isinstance(result, str) else 0,
                latency_ms=total_time,
            )
            
        except Exception as e:
            yield StreamEvent(
                type=StreamEventType.ERROR,
                data=str(e),
            )
    
    return wrapper


# =============================================================================
# Async Stream Builder
# =============================================================================

class StreamBuilder:
    """
    Fluent builder for streams.
    
    Usage:
        >>> result = await (
        ...     StreamBuilder(source)
        ...     .filter(lambda e: e.type == StreamEventType.TOKEN)
        ...     .transform(uppercase_transform)
        ...     .on_token(print)
        ...     .collect()
        ... )
    """
    
    def __init__(self, source: Union[StreamSource, AsyncGenerator]):
        self._source = source
        self._filters: List[Callable] = []
        self._transforms: List[Callable] = []
        self._callbacks: List[Callable] = []
    
    def filter(self, fn: Callable[[StreamEvent], bool]) -> "StreamBuilder":
        """Add a filter."""
        self._filters.append(fn)
        return self
    
    def transform(self, fn: Callable[[StreamEvent], StreamEvent]) -> "StreamBuilder":
        """Add a transform."""
        self._transforms.append(fn)
        return self
    
    def on_token(self, fn: Callable[[str], None]) -> "StreamBuilder":
        """Add token callback."""
        async def callback(event: StreamEvent):
            if event.type == StreamEventType.TOKEN and event.token:
                fn(event.token)
        self._callbacks.append(callback)
        return self
    
    def on_complete(self, fn: Callable[[], None]) -> "StreamBuilder":
        """Add completion callback."""
        async def callback(event: StreamEvent):
            if event.type == StreamEventType.COMPLETE:
                fn()
        self._callbacks.append(callback)
        return self
    
    async def collect(self) -> str:
        """Collect stream into string."""
        return await stream_to_string(self._process())
    
    async def _process(self) -> AsyncGenerator[StreamEvent, None]:
        """Process the stream."""
        source = self._source
        
        if hasattr(source, "stream"):
            source = source.stream()
        
        async for event in source:
            # Apply filters
            if any(not f(event) for f in self._filters):
                continue
            
            # Apply transforms
            for transform in self._transforms:
                event = transform(event)
            
            # Call callbacks
            for callback in self._callbacks:
                await callback(event)
            
            yield event



"""
Enterprise Module - Minimal Code Patterns for Production AI Applications.

This module provides enterprise-grade decorators, factories, and blueprints
that enable developers to create complete AI applications with minimal code.

Features:
- Decorators: @agent, @workflow, @tool, @guardrail, @pipeline
- Factories: AgentFactory, WorkflowFactory, SDLCFactory
- Blueprints: Pre-built agents for common SDLC phases
- Adapters: Unified interfaces for cloud providers
- Presets: Enterprise configurations for common scenarios

Quick Start:
    >>> from agenticaiframework.enterprise import create_sdlc_pipeline, agent, workflow
    >>> 
    >>> # Create complete SDLC pipeline in 3 lines
    >>> pipeline = create_sdlc_pipeline("my-project")
    >>> result = await pipeline.run("Build an e-commerce platform")
    >>> print(result.artifacts)
    
    >>> # Or use decorators for custom agents
    >>> @agent(role="analyst", model="gpt-4o")
    >>> class MyAnalyst:
    ...     async def analyze(self, data: str) -> dict:
    ...         return await self.invoke(f"Analyze: {data}")
"""

__version__ = "1.0.0"

# Decorators
from .decorators import (
    agent,
    workflow,
    tool,
    guardrail,
    pipeline,
    step,
    retry,
    timeout,
    trace,
    cache,
    validate,
    authorize,
)

# Factories
from .factories import (
    AgentFactory,
    WorkflowFactory,
    PipelineFactory,
    ToolFactory,
    create_agent,
    create_workflow,
    create_pipeline,
)

# Blueprints (Pre-built agents)
from .blueprints import (
    # SDLC Agents
    RequirementsAgent,
    DesignAgent,
    DevelopmentAgent,
    TestingAgent,
    SecurityAgent,
    DeploymentAgent,
    DocumentationAgent,
    ReviewAgent,
    # Other common agents
    AnalystAgent,
    ResearcherAgent,
    WriterAgent,
    # Blueprints helpers
    get_blueprint,
    list_blueprints,
)

# SDLC Pipeline (Complete solution)
from .sdlc import (
    SDLCPipeline,
    SDLCPhase,
    SDLCArtifact,
    SDLCConfig,
    SDLCResult,
    create_sdlc_pipeline,
)

# Adapters
from .adapters import (
    StorageAdapter,
    LLMAdapter,
    VectorDBAdapter,
    CacheAdapter,
    # Cloud-specific
    AzureAdapter,
    AWSAdapter,
    GCPAdapter,
    get_adapter,
)

# Presets
from .presets import (
    EnterprisePreset,
    StartupPreset,
    DevelopmentPreset,
    ProductionPreset,
    TestingPreset,
    MinimalPreset,
    load_preset,
    list_presets,
    auto_preset,
)

# Utils
from .utils import (
    AsyncContextMixin,
    StreamingResponse,
    stream_response,
    BatchProcessor,
    BatchResult,
    ProgressTracker,
    CancellationToken,
    retry_with_backoff,
    CircuitBreaker,
    RateLimiter,
    KeyedRateLimiter,
    ConnectionPool,
    AsyncCache,
    cached,
    StructuredLogger,
    HealthChecker,
    HealthStatus,
    MetricsCollector,
)

# Middleware
from .middleware import (
    Request,
    Response,
    Middleware,
    MiddlewarePipeline,
    LoggingMiddleware,
    MetricsMiddleware,
    AuthMiddleware,
    RateLimitMiddleware,
    CachingMiddleware,
    RetryMiddleware,
    TracingMiddleware,
    ValidationMiddleware,
    HooksManager,
    HookType,
    create_standard_pipeline,
    create_secure_pipeline,
    create_cached_pipeline,
)

# Events
from .events import (
    Event,
    EventType,
    EventBus,
    EventAggregator,
    EventStore,
    InMemoryEventStore,
    TypedEvent,
    global_event_bus,
    on_event,
    emit_event,
    emit_agent_started,
    emit_agent_completed,
    emit_agent_failed,
)

# Dependency Injection
from .di import (
    Container,
    Scope,
    Lifetime,
    ServiceCollection,
    ServiceProvider,
    inject,
    inject_async,
    singleton,
    transient,
    scoped,
    resolve,
    get_container,
    set_container,
    DependencyError,
)

# Plugins
from .plugins import (
    Plugin,
    PluginMetadata,
    PluginStatus,
    PluginRegistry,
    PluginLoader,
    PluginManager,
    plugin,
    get_plugin_registry,
    register_plugin,
    get_plugin,
    PluginError,
)

# Secrets
from .secrets import (
    Secret,
    SecretType,
    SecretBackend,
    EnvironmentBackend,
    InMemoryBackend,
    AzureKeyVaultBackend,
    SecretManager,
    get_secret_manager,
    get_secret,
    set_secret,
    rotate_secret,
    secret_from_env,
    mask_secret,
    generate_secret,
)

# Audit
from .audit import (
    AuditEntry,
    AuditAction,
    AuditSeverity,
    AuditBackend,
    InMemoryAuditBackend,
    FileAuditBackend,
    AuditLogger,
    get_audit_logger,
    audit_log,
    audit_track,
)

# RBAC
from .rbac import (
    Action,
    ResourceType,
    Permission,
    Role,
    Principal,
    Policy,
    AccessDecision,
    RBACManager,
    ROLE_ADMIN,
    ROLE_OPERATOR,
    ROLE_DEVELOPER,
    ROLE_VIEWER,
    ROLE_AGENT,
    get_current_principal,
    set_current_principal,
    PrincipalContext,
    get_rbac_manager,
    check_access,
    require_access,
)

# Templates
from .templates import (
    PromptTemplate,
    TemplateType,
    TemplateVariable,
    TemplateLibrary,
    TemplateBuilder,
    template,
    get_template_library,
    render_template,
    register_template,
    get_template,
)

# Feature Flags
from .feature_flags import (
    FeatureFlag,
    FlagType,
    RolloutStrategy,
    FlagVariant,
    FlagContext,
    FlagEvaluation,
    FlagBackend,
    InMemoryFlagBackend,
    FlagEvaluator,
    FeatureFlags,
    feature_flag,
    get_feature_flags,
    is_enabled,
    get_flag_value,
    get_variant,
)

# Multi-tenancy
from .multitenancy import (
    Tenant,
    TenantStatus,
    TenantTier,
    TenantQuota,
    TenantUsage,
    TenantStore,
    InMemoryTenantStore,
    TenantManager,
    TenantContext,
    TenantIsolation,
    TenantMiddleware,
    get_tenant_manager,
    get_current_tenant,
    require_tenant,
    check_quota,
    require_feature,
)

# Streaming
from .streaming import (
    StreamEventType,
    StreamEvent,
    StreamSource,
    LLMStreamSource,
    StreamProcessor,
    StreamAccumulator,
    StreamMultiplexer,
    SSEFormatter,
    JSONLFormatter,
    MarkdownFormatter,
    stream_to_string,
    stream_with_progress,
    stream_with_display,
    StreamBuilder,
)

# Cost Estimation
from .cost import (
    ModelPricing,
    TokenCounter,
    TiktokenCounter,
    ApproximateCounter,
    get_token_counter,
    UsageRecord,
    CostEstimator,
    CostEstimate,
    UsageTracker,
    UsageSummary,
    get_usage_tracker,
    set_usage_tracker,
    track_usage,
    get_usage_summary,
    estimate_cost,
    count_tokens,
)

# JSON Mode
from .json_mode import (
    python_type_to_json_schema,
    generate_json_schema,
    JSONExtractor,
    JSONRepairer,
    JSONResponse,
    JSONMode,
    json_output,
    TextWithConfidence,
    EntityExtraction,
    Classification,
    Sentiment,
    Summary,
    Translation,
    CodeGeneration,
    StructuredAnswer,
    create_json_mode,
    extract_json,
    repair_json,
    schema_from_model,
)

# State Management
from .state import (
    StateStatus,
    StateMetadata,
    StateEntry,
    StateStore,
    InMemoryStateStore,
    FileStateStore,
    Checkpoint,
    CheckpointManager,
    StateManager,
    ScopedStateManager,
    StateTransaction,
    get_state_manager,
    set_state_manager,
    get_state,
    set_state,
    delete_state,
    stateful,
)

# Parallel Execution
from .parallel import (
    TaskStatus,
    TaskPriority,
    ParallelTask,
    TaskResult,
    ResourcePool,
    ResourceContext,
    DAGExecutor,
    parallel_map,
    parallel_batch,
    parallel_step,
    collect_parallel_steps,
    ExecutionProgress,
    ProgressTracker as ParallelProgressTracker,
    create_dag,
    task,
    run_parallel,
    run_with_timeout,
)

# Distributed Tracing (OpenTelemetry)
from .tracing_otel import (
    SpanKind,
    SpanStatus,
    SpanContext,
    SpanEvent,
    Span,
    SpanExporter,
    ConsoleExporter,
    InMemoryExporter,
    OTLPExporter,
    Tracer,
    get_tracer,
    set_tracer,
    configure_tracing,
    trace as otel_trace,
    trace_agent,
    inject_context,
    extract_context,
    with_context,
    AgentTracer,
    start_span,
    end_span,
    current_trace_id,
    current_span_id,
)

# Metrics Export
from .metrics import (
    MetricType,
    MetricLabel,
    MetricValue,
    Metric,
    Counter,
    Gauge,
    Histogram,
    Timer,
    MetricsRegistry,
    MetricsExporter,
    PrometheusExporter,
    StatsDExporter,
    get_metrics_registry,
    set_metrics_registry,
    AgentMetrics,
    with_metrics,
    start_metrics_server,
)

# Workflow DSL
from .dsl import (
    NodeType,
    ConditionOperator,
    Condition,
    WorkflowNode,
    AgentNode,
    ToolNode,
    ConditionNode,
    LoopNode,
    ParallelNode,
    HumanNode,
    TransformNode,
    WorkflowDefinition,
    WorkflowBuilder,
    ConditionBuilder,
    ParallelBuilder,
    LoopBuilder,
    WorkflowParser,
    workflow as dsl_workflow,
    parse_workflow,
)

# Agent Composition
from .composition import (
    ComponentStatus,
    ComponentMetadata,
    Component,
    MemoryTrait,
    ToolsTrait,
    ObservableTrait,
    ReflectionTrait,
    ComposableAgent,
    InputValidator,
    OutputFormatter,
    LoggingComponent,
    TransformComponent,
    FilterComponent,
    CacheComponent,
    AgentPipeline,
    ComposableAgentFactory,
    AgentBuilder,
    component,
    composable,
    create_pipeline as create_agent_pipeline,
    compose,
)

# API Generator
from .api_gen import (
    SchemaObject,
    ParameterObject,
    RequestBodyObject,
    ResponseObject,
    OperationObject,
    PathItemObject,
    OpenAPISpec,
    SchemaConverter,
    APIGenerator,
    api_endpoint,
    create_fastapi_app,
    register_agent_routes,
    generate_openapi,
    save_openapi_spec,
)

__all__ = [
    # Decorators
    "agent",
    "workflow",
    "tool",
    "guardrail",
    "pipeline",
    "step",
    "retry",
    "timeout",
    "trace",
    "cache",
    "validate",
    "authorize",
    # Factories
    "AgentFactory",
    "WorkflowFactory",
    "PipelineFactory",
    "ToolFactory",
    "create_agent",
    "create_workflow",
    "create_pipeline",
    # Blueprints
    "RequirementsAgent",
    "DesignAgent",
    "DevelopmentAgent",
    "TestingAgent",
    "SecurityAgent",
    "DeploymentAgent",
    "DocumentationAgent",
    "ReviewAgent",
    "AnalystAgent",
    "ResearcherAgent",
    "WriterAgent",
    "get_blueprint",
    "list_blueprints",
    # SDLC
    "SDLCPipeline",
    "SDLCPhase",
    "SDLCArtifact",
    "SDLCConfig",
    "SDLCResult",
    "create_sdlc_pipeline",
    # Adapters
    "StorageAdapter",
    "LLMAdapter",
    "VectorDBAdapter",
    "CacheAdapter",
    "AzureAdapter",
    "AWSAdapter",
    "GCPAdapter",
    "get_adapter",
    # Presets
    "EnterprisePreset",
    "StartupPreset",
    "DevelopmentPreset",
    "ProductionPreset",
    "TestingPreset",
    "MinimalPreset",
    "load_preset",
    "list_presets",
    "auto_preset",
    # Utils
    "AsyncContextMixin",
    "StreamingResponse",
    "stream_response",
    "BatchProcessor",
    "BatchResult",
    "ProgressTracker",
    "CancellationToken",
    "retry_with_backoff",
    "CircuitBreaker",
    "RateLimiter",
    "KeyedRateLimiter",
    "ConnectionPool",
    "AsyncCache",
    "cached",
    "StructuredLogger",
    "HealthChecker",
    "HealthStatus",
    "MetricsCollector",
    # Middleware
    "Request",
    "Response",
    "Middleware",
    "MiddlewarePipeline",
    "LoggingMiddleware",
    "MetricsMiddleware",
    "AuthMiddleware",
    "RateLimitMiddleware",
    "CachingMiddleware",
    "RetryMiddleware",
    "TracingMiddleware",
    "ValidationMiddleware",
    "HooksManager",
    "HookType",
    "create_standard_pipeline",
    "create_secure_pipeline",
    "create_cached_pipeline",
    # Events
    "Event",
    "EventType",
    "EventBus",
    "EventAggregator",
    "EventStore",
    "InMemoryEventStore",
    "TypedEvent",
    "global_event_bus",
    "on_event",
    "emit_event",
    "emit_agent_started",
    "emit_agent_completed",
    "emit_agent_failed",
    # Dependency Injection
    "Container",
    "Scope",
    "Lifetime",
    "ServiceCollection",
    "ServiceProvider",
    "inject",
    "inject_async",
    "singleton",
    "transient",
    "scoped",
    "resolve",
    "get_container",
    "set_container",
    "DependencyError",
    # Plugins
    "Plugin",
    "PluginMetadata",
    "PluginStatus",
    "PluginRegistry",
    "PluginLoader",
    "PluginManager",
    "plugin",
    "get_plugin_registry",
    "register_plugin",
    "get_plugin",
    "PluginError",
    # Secrets
    "Secret",
    "SecretType",
    "SecretBackend",
    "EnvironmentBackend",
    "InMemoryBackend",
    "AzureKeyVaultBackend",
    "SecretManager",
    "get_secret_manager",
    "get_secret",
    "set_secret",
    "rotate_secret",
    "secret_from_env",
    "mask_secret",
    "generate_secret",
    # Audit
    "AuditEntry",
    "AuditAction",
    "AuditSeverity",
    "AuditBackend",
    "InMemoryAuditBackend",
    "FileAuditBackend",
    "AuditLogger",
    "get_audit_logger",
    "audit_log",
    "audit_track",
    # RBAC
    "Action",
    "ResourceType",
    "Permission",
    "Role",
    "Principal",
    "Policy",
    "AccessDecision",
    "RBACManager",
    "ROLE_ADMIN",
    "ROLE_OPERATOR",
    "ROLE_DEVELOPER",
    "ROLE_VIEWER",
    "ROLE_AGENT",
    "get_current_principal",
    "set_current_principal",
    "PrincipalContext",
    "get_rbac_manager",
    "check_access",
    "require_access",
    # Templates
    "PromptTemplate",
    "TemplateType",
    "TemplateVariable",
    "TemplateLibrary",
    "TemplateBuilder",
    "template",
    "get_template_library",
    "render_template",
    "register_template",
    "get_template",
    # Feature Flags
    "FeatureFlag",
    "FlagType",
    "RolloutStrategy",
    "FlagVariant",
    "FlagContext",
    "FlagEvaluation",
    "FlagBackend",
    "InMemoryFlagBackend",
    "FlagEvaluator",
    "FeatureFlags",
    "feature_flag",
    "get_feature_flags",
    "is_enabled",
    "get_flag_value",
    "get_variant",
    # Multi-tenancy
    "Tenant",
    "TenantStatus",
    "TenantTier",
    "TenantQuota",
    "TenantUsage",
    "TenantStore",
    "InMemoryTenantStore",
    "TenantManager",
    "TenantContext",
    "TenantIsolation",
    "TenantMiddleware",
    "get_tenant_manager",
    "get_current_tenant",
    "require_tenant",
    "check_quota",
    "require_feature",
    # Streaming
    "StreamEventType",
    "StreamEvent",
    "StreamSource",
    "LLMStreamSource",
    "StreamProcessor",
    "StreamAccumulator",
    "StreamMultiplexer",
    "SSEFormatter",
    "JSONLFormatter",
    "MarkdownFormatter",
    "stream_to_string",
    "stream_with_progress",
    "stream_with_display",
    "StreamBuilder",
    # Cost Estimation
    "ModelPricing",
    "TokenCounter",
    "TiktokenCounter",
    "ApproximateCounter",
    "get_token_counter",
    "UsageRecord",
    "CostEstimator",
    "CostEstimate",
    "UsageTracker",
    "UsageSummary",
    "get_usage_tracker",
    "set_usage_tracker",
    "track_usage",
    "get_usage_summary",
    "estimate_cost",
    "count_tokens",
    # JSON Mode
    "python_type_to_json_schema",
    "generate_json_schema",
    "JSONExtractor",
    "JSONRepairer",
    "JSONResponse",
    "JSONMode",
    "json_output",
    "TextWithConfidence",
    "EntityExtraction",
    "Classification",
    "Sentiment",
    "Summary",
    "Translation",
    "CodeGeneration",
    "StructuredAnswer",
    "create_json_mode",
    "extract_json",
    "repair_json",
    "schema_from_model",
    # State Management
    "StateStatus",
    "StateMetadata",
    "StateEntry",
    "StateStore",
    "InMemoryStateStore",
    "FileStateStore",
    "Checkpoint",
    "CheckpointManager",
    "StateManager",
    "ScopedStateManager",
    "StateTransaction",
    "get_state_manager",
    "set_state_manager",
    "get_state",
    "set_state",
    "delete_state",
    "stateful",
    # Parallel Execution
    "TaskStatus",
    "TaskPriority",
    "ParallelTask",
    "TaskResult",
    "ResourcePool",
    "ResourceContext",
    "DAGExecutor",
    "parallel_map",
    "parallel_batch",
    "parallel_step",
    "collect_parallel_steps",
    "ExecutionProgress",
    "ParallelProgressTracker",
    "create_dag",
    "task",
    "run_parallel",
    "run_with_timeout",
    # Distributed Tracing
    "SpanKind",
    "SpanStatus",
    "SpanContext",
    "SpanEvent",
    "Span",
    "SpanExporter",
    "ConsoleExporter",
    "InMemoryExporter",
    "OTLPExporter",
    "Tracer",
    "get_tracer",
    "set_tracer",
    "configure_tracing",
    "otel_trace",
    "trace_agent",
    "inject_context",
    "extract_context",
    "with_context",
    "AgentTracer",
    "start_span",
    "end_span",
    "current_trace_id",
    "current_span_id",
    # Metrics Export
    "MetricType",
    "MetricLabel",
    "MetricValue",
    "Metric",
    "Counter",
    "Gauge",
    "Histogram",
    "Timer",
    "MetricsRegistry",
    "MetricsExporter",
    "PrometheusExporter",
    "StatsDExporter",
    "get_metrics_registry",
    "set_metrics_registry",
    "AgentMetrics",
    "with_metrics",
    "start_metrics_server",
    # Workflow DSL
    "NodeType",
    "ConditionOperator",
    "Condition",
    "WorkflowNode",
    "AgentNode",
    "ToolNode",
    "ConditionNode",
    "LoopNode",
    "ParallelNode",
    "HumanNode",
    "TransformNode",
    "WorkflowDefinition",
    "WorkflowBuilder",
    "ConditionBuilder",
    "ParallelBuilder",
    "LoopBuilder",
    "WorkflowParser",
    "dsl_workflow",
    "parse_workflow",
    # Agent Composition
    "ComponentStatus",
    "ComponentMetadata",
    "Component",
    "MemoryTrait",
    "ToolsTrait",
    "ObservableTrait",
    "ReflectionTrait",
    "ComposableAgent",
    "InputValidator",
    "OutputFormatter",
    "LoggingComponent",
    "TransformComponent",
    "FilterComponent",
    "CacheComponent",
    "AgentPipeline",
    "ComposableAgentFactory",
    "AgentBuilder",
    "component",
    "composable",
    "create_agent_pipeline",
    "compose",
    # API Generator
    "SchemaObject",
    "ParameterObject",
    "RequestBodyObject",
    "ResponseObject",
    "OperationObject",
    "PathItemObject",
    "OpenAPISpec",
    "SchemaConverter",
    "APIGenerator",
    "api_endpoint",
    "create_fastapi_app",
    "register_agent_routes",
    "generate_openapi",
    "save_openapi_spec",
]



"""
Enterprise Chain Module.

Provides LLM chain composition, sequential processing,
and chain-of-thought patterns for agent operations.

Example:
    # Build a chain
    chain = (
        Chain()
        .prompt("Analyze this: {input}")
        .llm(openai_client)
        .parse(json_parser)
        .transform(lambda x: x["result"])
    )
    
    result = await chain.run(input="some text")
    
    # Decorators
    @chain_step(name="preprocess")
    async def preprocess(data):
        ...
"""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
)
from datetime import datetime
from functools import wraps
from enum import Enum
import logging
import time
import traceback

logger = logging.getLogger(__name__)

T = TypeVar('T')
R = TypeVar('R')
InputType = TypeVar('InputType')
OutputType = TypeVar('OutputType')


class ChainError(Exception):
    """Chain execution error."""
    pass


class ChainStepError(ChainError):
    """Error in a chain step."""
    
    def __init__(
        self,
        message: str,
        step_name: str,
        step_index: int,
        original: Optional[Exception] = None,
    ):
        super().__init__(message)
        self.step_name = step_name
        self.step_index = step_index
        self.original = original


class ChainStatus(str, Enum):
    """Chain execution status."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"


@dataclass
class StepResult:
    """Result of a chain step."""
    step_name: str
    step_index: int
    input_data: Any
    output_data: Any
    duration_ms: float
    success: bool
    error: Optional[str] = None
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "step_name": self.step_name,
            "step_index": self.step_index,
            "duration_ms": round(self.duration_ms, 2),
            "success": self.success,
            "error": self.error,
        }


@dataclass
class ChainResult:
    """Result of chain execution."""
    status: ChainStatus
    output: Any
    steps: List[StepResult] = field(default_factory=list)
    total_duration_ms: float = 0.0
    error: Optional[str] = None
    
    @property
    def success(self) -> bool:
        """Check if chain succeeded."""
        return self.status == ChainStatus.COMPLETED
    
    @property
    def step_count(self) -> int:
        """Get number of steps executed."""
        return len(self.steps)
    
    def to_dict(self) -> Dict[str, Any]:
        return {
            "status": self.status.value,
            "output": str(self.output)[:200] if self.output else None,
            "steps": [s.to_dict() for s in self.steps],
            "total_duration_ms": round(self.total_duration_ms, 2),
            "error": self.error,
        }


class ChainStep(ABC, Generic[T, R]):
    """Abstract chain step."""
    
    @property
    @abstractmethod
    def name(self) -> str:
        """Get step name."""
        pass
    
    @abstractmethod
    async def execute(self, input_data: T) -> R:
        """Execute the step."""
        pass


class FunctionStep(ChainStep[T, R]):
    """Step that executes a function."""
    
    def __init__(
        self,
        func: Callable[[T], Union[R, Awaitable[R]]],
        step_name: Optional[str] = None,
    ):
        self._func = func
        self._name = step_name or func.__name__
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: T) -> R:
        """Execute the function."""
        if asyncio.iscoroutinefunction(self._func):
            return await self._func(input_data)
        return self._func(input_data)


class PromptStep(ChainStep[Dict[str, Any], str]):
    """Step that formats a prompt template."""
    
    def __init__(
        self,
        template: str,
        step_name: str = "prompt",
    ):
        self._template = template
        self._name = step_name
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: Dict[str, Any]) -> str:
        """Format the prompt template."""
        if isinstance(input_data, dict):
            return self._template.format(**input_data)
        return self._template.format(input=input_data)


class LLMStep(ChainStep[str, str]):
    """Step that calls an LLM."""
    
    def __init__(
        self,
        llm_client: Any,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        step_name: str = "llm",
    ):
        self._client = llm_client
        self._model = model
        self._temperature = temperature
        self._max_tokens = max_tokens
        self._name = step_name
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: str) -> str:
        """Call the LLM."""
        # Generic LLM call pattern
        if hasattr(self._client, 'chat'):
            response = await self._client.chat.completions.create(
                model=self._model,
                messages=[{"role": "user", "content": input_data}],
                temperature=self._temperature,
                max_tokens=self._max_tokens,
            )
            return response.choices[0].message.content
        
        if hasattr(self._client, 'generate'):
            response = await self._client.generate(
                prompt=input_data,
                temperature=self._temperature,
                max_tokens=self._max_tokens,
            )
            return response
        
        raise ChainError(f"Unsupported LLM client: {type(self._client)}")


class ParseStep(ChainStep[str, Any]):
    """Step that parses output."""
    
    def __init__(
        self,
        parser: Callable[[str], Any],
        step_name: str = "parse",
    ):
        self._parser = parser
        self._name = step_name
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: str) -> Any:
        """Parse the input."""
        if asyncio.iscoroutinefunction(self._parser):
            return await self._parser(input_data)
        return self._parser(input_data)


class BranchStep(ChainStep[T, R]):
    """Step that branches based on a condition."""
    
    def __init__(
        self,
        condition: Callable[[T], bool],
        if_true: ChainStep[T, R],
        if_false: ChainStep[T, R],
        step_name: str = "branch",
    ):
        self._condition = condition
        self._if_true = if_true
        self._if_false = if_false
        self._name = step_name
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: T) -> R:
        """Execute the appropriate branch."""
        if asyncio.iscoroutinefunction(self._condition):
            result = await self._condition(input_data)
        else:
            result = self._condition(input_data)
        
        if result:
            return await self._if_true.execute(input_data)
        return await self._if_false.execute(input_data)


class MapStep(ChainStep[List[T], List[R]]):
    """Step that maps over a list."""
    
    def __init__(
        self,
        step: ChainStep[T, R],
        parallel: bool = True,
        step_name: str = "map",
    ):
        self._step = step
        self._parallel = parallel
        self._name = step_name
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: List[T]) -> List[R]:
        """Map the step over the list."""
        if self._parallel:
            tasks = [self._step.execute(item) for item in input_data]
            return await asyncio.gather(*tasks)
        
        results = []
        for item in input_data:
            result = await self._step.execute(item)
            results.append(result)
        return results


class ReduceStep(ChainStep[List[T], R]):
    """Step that reduces a list to a single value."""
    
    def __init__(
        self,
        reducer: Callable[[R, T], R],
        initial: R,
        step_name: str = "reduce",
    ):
        self._reducer = reducer
        self._initial = initial
        self._name = step_name
    
    @property
    def name(self) -> str:
        return self._name
    
    async def execute(self, input_data: List[T]) -> R:
        """Reduce the list."""
        result = self._initial
        
        for item in input_data:
            if asyncio.iscoroutinefunction(self._reducer):
                result = await self._reducer(result, item)
            else:
                result = self._reducer(result, item)
        
        return result


class Chain(Generic[InputType, OutputType]):
    """
    LLM chain for sequential processing.
    """
    
    def __init__(
        self,
        name: str = "chain",
        steps: Optional[List[ChainStep]] = None,
    ):
        """
        Initialize chain.
        
        Args:
            name: Chain name
            steps: Initial steps
        """
        self._name = name
        self._steps: List[ChainStep] = list(steps) if steps else []
        self._on_step: Optional[Callable[[StepResult], None]] = None
        self._on_error: Optional[Callable[[ChainStepError], None]] = None
    
    @property
    def name(self) -> str:
        return self._name
    
    def add(self, step: ChainStep) -> 'Chain':
        """Add a step to the chain."""
        self._steps.append(step)
        return self
    
    def prompt(
        self,
        template: str,
        name: str = "prompt",
    ) -> 'Chain':
        """Add a prompt formatting step."""
        return self.add(PromptStep(template, name))
    
    def llm(
        self,
        client: Any,
        model: Optional[str] = None,
        temperature: float = 0.7,
        max_tokens: Optional[int] = None,
        name: str = "llm",
    ) -> 'Chain':
        """Add an LLM call step."""
        return self.add(LLMStep(client, model, temperature, max_tokens, name))
    
    def parse(
        self,
        parser: Callable[[str], Any],
        name: str = "parse",
    ) -> 'Chain':
        """Add a parsing step."""
        return self.add(ParseStep(parser, name))
    
    def transform(
        self,
        func: Callable[[Any], Any],
        name: Optional[str] = None,
    ) -> 'Chain':
        """Add a transformation step."""
        return self.add(FunctionStep(func, name))
    
    def branch(
        self,
        condition: Callable[[Any], bool],
        if_true: ChainStep,
        if_false: ChainStep,
        name: str = "branch",
    ) -> 'Chain':
        """Add a branching step."""
        return self.add(BranchStep(condition, if_true, if_false, name))
    
    def map(
        self,
        step: ChainStep,
        parallel: bool = True,
        name: str = "map",
    ) -> 'Chain':
        """Add a map step."""
        return self.add(MapStep(step, parallel, name))
    
    def on_step(
        self,
        callback: Callable[[StepResult], None],
    ) -> 'Chain':
        """Set step callback."""
        self._on_step = callback
        return self
    
    def on_error(
        self,
        callback: Callable[[ChainStepError], None],
    ) -> 'Chain':
        """Set error callback."""
        self._on_error = callback
        return self
    
    async def run(
        self,
        input_data: Any = None,
        **kwargs: Any,
    ) -> ChainResult:
        """
        Run the chain.
        
        Args:
            input_data: Initial input (or pass as kwargs)
            **kwargs: Named inputs
            
        Returns:
            Chain result
        """
        if input_data is None and kwargs:
            input_data = kwargs
        
        start_time = time.time()
        steps_results = []
        current_data = input_data
        
        for i, step in enumerate(self._steps):
            step_start = time.time()
            
            try:
                output = await step.execute(current_data)
                step_duration = (time.time() - step_start) * 1000
                
                result = StepResult(
                    step_name=step.name,
                    step_index=i,
                    input_data=current_data,
                    output_data=output,
                    duration_ms=step_duration,
                    success=True,
                )
                
                steps_results.append(result)
                
                if self._on_step:
                    self._on_step(result)
                
                current_data = output
                
            except Exception as e:
                step_duration = (time.time() - step_start) * 1000
                
                error = ChainStepError(
                    str(e),
                    step_name=step.name,
                    step_index=i,
                    original=e,
                )
                
                if self._on_error:
                    self._on_error(error)
                
                result = StepResult(
                    step_name=step.name,
                    step_index=i,
                    input_data=current_data,
                    output_data=None,
                    duration_ms=step_duration,
                    success=False,
                    error=str(e),
                )
                
                steps_results.append(result)
                
                return ChainResult(
                    status=ChainStatus.FAILED,
                    output=None,
                    steps=steps_results,
                    total_duration_ms=(time.time() - start_time) * 1000,
                    error=str(error),
                )
        
        return ChainResult(
            status=ChainStatus.COMPLETED,
            output=current_data,
            steps=steps_results,
            total_duration_ms=(time.time() - start_time) * 1000,
        )
    
    async def stream(
        self,
        input_data: Any = None,
        **kwargs: Any,
    ) -> AsyncIterator[StepResult]:
        """
        Stream chain execution.
        
        Yields step results as they complete.
        """
        if input_data is None and kwargs:
            input_data = kwargs
        
        current_data = input_data
        
        for i, step in enumerate(self._steps):
            step_start = time.time()
            
            try:
                output = await step.execute(current_data)
                step_duration = (time.time() - step_start) * 1000
                
                result = StepResult(
                    step_name=step.name,
                    step_index=i,
                    input_data=current_data,
                    output_data=output,
                    duration_ms=step_duration,
                    success=True,
                )
                
                yield result
                current_data = output
                
            except Exception as e:
                step_duration = (time.time() - step_start) * 1000
                
                result = StepResult(
                    step_name=step.name,
                    step_index=i,
                    input_data=current_data,
                    output_data=None,
                    duration_ms=step_duration,
                    success=False,
                    error=str(e),
                )
                
                yield result
                break
    
    def __or__(self, other: ChainStep) -> 'Chain':
        """Pipe operator to add steps."""
        return self.add(other)
    
    def __rshift__(self, func: Callable) -> 'Chain':
        """Right shift to add transform."""
        return self.transform(func)


class ParallelChain:
    """
    Execute multiple chains in parallel.
    """
    
    def __init__(self, chains: Optional[List[Chain]] = None):
        self._chains = list(chains) if chains else []
    
    def add(self, chain: Chain) -> 'ParallelChain':
        """Add a chain."""
        self._chains.append(chain)
        return self
    
    async def run(
        self,
        input_data: Any,
        merge: Optional[Callable[[List[Any]], Any]] = None,
    ) -> List[ChainResult]:
        """Run all chains in parallel."""
        tasks = [chain.run(input_data) for chain in self._chains]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        
        chain_results = []
        for result in results:
            if isinstance(result, Exception):
                chain_results.append(ChainResult(
                    status=ChainStatus.FAILED,
                    output=None,
                    error=str(result),
                ))
            else:
                chain_results.append(result)
        
        return chain_results


class SequentialChain:
    """
    Execute chains sequentially, passing output to next chain.
    """
    
    def __init__(self, chains: Optional[List[Chain]] = None):
        self._chains = list(chains) if chains else []
    
    def add(self, chain: Chain) -> 'SequentialChain':
        """Add a chain."""
        self._chains.append(chain)
        return self
    
    async def run(self, input_data: Any) -> ChainResult:
        """Run chains sequentially."""
        current_data = input_data
        all_steps = []
        start_time = time.time()
        
        for chain in self._chains:
            result = await chain.run(current_data)
            all_steps.extend(result.steps)
            
            if not result.success:
                return ChainResult(
                    status=ChainStatus.FAILED,
                    output=None,
                    steps=all_steps,
                    total_duration_ms=(time.time() - start_time) * 1000,
                    error=result.error,
                )
            
            current_data = result.output
        
        return ChainResult(
            status=ChainStatus.COMPLETED,
            output=current_data,
            steps=all_steps,
            total_duration_ms=(time.time() - start_time) * 1000,
        )


def chain_step(
    name: Optional[str] = None,
) -> Callable:
    """
    Decorator to create a chain step from a function.
    
    Example:
        @chain_step(name="preprocess")
        async def preprocess(data):
            return processed_data
    """
    def decorator(func: Callable) -> FunctionStep:
        step_name = name or func.__name__
        return FunctionStep(func, step_name)
    
    return decorator


def create_chain(
    *steps: Union[ChainStep, Callable],
    name: str = "chain",
) -> Chain:
    """
    Factory function to create a chain.
    
    Example:
        chain = create_chain(
            preprocess,
            analyze,
            format_output,
            name="analysis_chain",
        )
    """
    chain = Chain(name=name)
    
    for step in steps:
        if isinstance(step, ChainStep):
            chain.add(step)
        elif callable(step):
            chain.add(FunctionStep(step))
    
    return chain


__all__ = [
    # Exceptions
    "ChainError",
    "ChainStepError",
    # Enums
    "ChainStatus",
    # Data classes
    "StepResult",
    "ChainResult",
    # Steps
    "ChainStep",
    "FunctionStep",
    "PromptStep",
    "LLMStep",
    "ParseStep",
    "BranchStep",
    "MapStep",
    "ReduceStep",
    # Chains
    "Chain",
    "ParallelChain",
    "SequentialChain",
    # Decorators
    "chain_step",
    # Factory
    "create_chain",
]



"""
Enterprise Incident Manager Module.

Incident tracking, escalation workflows,
on-call management, and incident response.

Example:
    # Create incident manager
    incidents = create_incident_manager()
    
    # Create incident
    incident = await incidents.create(
        title="Database connection failure",
        severity=Severity.HIGH,
        affected_services=["api", "web"],
    )
    
    # Update status
    await incidents.update_status(
        incident_id=incident.id,
        status=IncidentStatus.INVESTIGATING,
    )
    
    # Add timeline entry
    await incidents.add_timeline_entry(
        incident_id=incident.id,
        message="Root cause identified: connection pool exhausted",
    )
    
    # Resolve incident
    await incidents.resolve(
        incident_id=incident.id,
        resolution="Increased connection pool size",
    )
"""

from __future__ import annotations

import asyncio
import logging
import time
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

logger = logging.getLogger(__name__)


class IncidentError(Exception):
    """Incident error."""
    pass


class Severity(str, Enum):
    """Incident severity."""
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


class IncidentStatus(str, Enum):
    """Incident status."""
    OPEN = "open"
    ACKNOWLEDGED = "acknowledged"
    INVESTIGATING = "investigating"
    IDENTIFIED = "identified"
    MONITORING = "monitoring"
    RESOLVED = "resolved"
    CLOSED = "closed"


class EscalationLevel(str, Enum):
    """Escalation level."""
    L1 = "L1"
    L2 = "L2"
    L3 = "L3"
    MANAGEMENT = "management"
    EXECUTIVE = "executive"


class TimelineEntryType(str, Enum):
    """Timeline entry type."""
    STATUS_CHANGE = "status_change"
    COMMENT = "comment"
    ESCALATION = "escalation"
    ASSIGNMENT = "assignment"
    ACTION = "action"
    NOTIFICATION = "notification"


@dataclass
class TimelineEntry:
    """Incident timeline entry."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    entry_type: TimelineEntryType = TimelineEntryType.COMMENT
    message: str = ""
    author: str = ""
    created_at: datetime = field(default_factory=datetime.utcnow)
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Responder:
    """Incident responder."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    email: str = ""
    role: str = ""
    escalation_level: EscalationLevel = EscalationLevel.L1
    on_call: bool = False
    notification_methods: List[str] = field(default_factory=list)


@dataclass
class EscalationPolicy:
    """Escalation policy."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    levels: Dict[EscalationLevel, List[str]] = field(default_factory=dict)
    escalation_timeout: int = 30  # minutes
    auto_escalate: bool = True
    notify_all_levels: bool = False


@dataclass
class Incident:
    """Incident."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    title: str = ""
    description: str = ""
    severity: Severity = Severity.MEDIUM
    status: IncidentStatus = IncidentStatus.OPEN
    affected_services: List[str] = field(default_factory=list)
    tags: Dict[str, str] = field(default_factory=dict)
    assignee: Optional[str] = None
    escalation_level: EscalationLevel = EscalationLevel.L1
    timeline: List[TimelineEntry] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)
    acknowledged_at: Optional[datetime] = None
    resolved_at: Optional[datetime] = None
    closed_at: Optional[datetime] = None
    resolution: str = ""
    root_cause: str = ""
    impact: str = ""
    customer_impact: bool = False
    public: bool = False
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    @property
    def duration(self) -> Optional[timedelta]:
        if self.resolved_at:
            return self.resolved_at - self.created_at
        return datetime.utcnow() - self.created_at
    
    @property
    def time_to_acknowledge(self) -> Optional[timedelta]:
        if self.acknowledged_at:
            return self.acknowledged_at - self.created_at
        return None
    
    @property
    def time_to_resolve(self) -> Optional[timedelta]:
        if self.resolved_at:
            return self.resolved_at - self.created_at
        return None


@dataclass
class IncidentStats:
    """Incident statistics."""
    total: int = 0
    open: int = 0
    resolved: int = 0
    by_severity: Dict[str, int] = field(default_factory=dict)
    by_status: Dict[str, int] = field(default_factory=dict)
    avg_time_to_resolve: Optional[float] = None
    avg_time_to_acknowledge: Optional[float] = None


@dataclass
class PostMortem:
    """Incident post-mortem."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    incident_id: str = ""
    title: str = ""
    summary: str = ""
    timeline: List[TimelineEntry] = field(default_factory=list)
    root_cause: str = ""
    contributing_factors: List[str] = field(default_factory=list)
    lessons_learned: List[str] = field(default_factory=list)
    action_items: List[Dict[str, Any]] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)
    author: str = ""


# Incident store
class IncidentStore(ABC):
    """Incident storage."""
    
    @abstractmethod
    async def save(self, incident: Incident) -> None:
        pass
    
    @abstractmethod
    async def get(self, incident_id: str) -> Optional[Incident]:
        pass
    
    @abstractmethod
    async def list_incidents(
        self,
        status: Optional[List[IncidentStatus]] = None,
        severity: Optional[List[Severity]] = None,
        limit: int = 100,
    ) -> List[Incident]:
        pass
    
    @abstractmethod
    async def delete(self, incident_id: str) -> bool:
        pass


class InMemoryIncidentStore(IncidentStore):
    """In-memory incident store."""
    
    def __init__(self):
        self._incidents: Dict[str, Incident] = {}
    
    async def save(self, incident: Incident) -> None:
        self._incidents[incident.id] = incident
    
    async def get(self, incident_id: str) -> Optional[Incident]:
        return self._incidents.get(incident_id)
    
    async def list_incidents(
        self,
        status: Optional[List[IncidentStatus]] = None,
        severity: Optional[List[Severity]] = None,
        limit: int = 100,
    ) -> List[Incident]:
        results = []
        
        for incident in self._incidents.values():
            if status and incident.status not in status:
                continue
            if severity and incident.severity not in severity:
                continue
            results.append(incident)
        
        results.sort(key=lambda i: i.created_at, reverse=True)
        return results[:limit]
    
    async def delete(self, incident_id: str) -> bool:
        if incident_id in self._incidents:
            del self._incidents[incident_id]
            return True
        return False


# Responder store
class ResponderStore(ABC):
    """Responder storage."""
    
    @abstractmethod
    async def save(self, responder: Responder) -> None:
        pass
    
    @abstractmethod
    async def get(self, responder_id: str) -> Optional[Responder]:
        pass
    
    @abstractmethod
    async def get_on_call(self, level: EscalationLevel) -> List[Responder]:
        pass
    
    @abstractmethod
    async def list_all(self) -> List[Responder]:
        pass


class InMemoryResponderStore(ResponderStore):
    """In-memory responder store."""
    
    def __init__(self):
        self._responders: Dict[str, Responder] = {}
    
    async def save(self, responder: Responder) -> None:
        self._responders[responder.id] = responder
    
    async def get(self, responder_id: str) -> Optional[Responder]:
        return self._responders.get(responder_id)
    
    async def get_on_call(self, level: EscalationLevel) -> List[Responder]:
        return [
            r for r in self._responders.values()
            if r.on_call and r.escalation_level == level
        ]
    
    async def list_all(self) -> List[Responder]:
        return list(self._responders.values())


# Notifier
class Notifier(ABC):
    """Notification interface."""
    
    @abstractmethod
    async def notify(
        self,
        incident: Incident,
        responders: List[Responder],
        message: str,
    ) -> None:
        pass


class LogNotifier(Notifier):
    """Log-based notifier for testing."""
    
    async def notify(
        self,
        incident: Incident,
        responders: List[Responder],
        message: str,
    ) -> None:
        for responder in responders:
            logger.info(
                f"Notification to {responder.name} ({responder.email}): "
                f"[{incident.severity.value.upper()}] {incident.title} - {message}"
            )


# Incident manager
class IncidentManager:
    """Incident manager."""
    
    def __init__(
        self,
        incident_store: Optional[IncidentStore] = None,
        responder_store: Optional[ResponderStore] = None,
        notifier: Optional[Notifier] = None,
    ):
        self._incident_store = incident_store or InMemoryIncidentStore()
        self._responder_store = responder_store or InMemoryResponderStore()
        self._notifier = notifier or LogNotifier()
        self._escalation_policies: Dict[str, EscalationPolicy] = {}
        self._listeners: List[Callable] = []
    
    async def create(
        self,
        title: str,
        severity: Union[str, Severity] = Severity.MEDIUM,
        description: str = "",
        affected_services: Optional[List[str]] = None,
        tags: Optional[Dict[str, str]] = None,
        customer_impact: bool = False,
        **metadata,
    ) -> Incident:
        """Create incident."""
        if isinstance(severity, str):
            severity = Severity(severity)
        
        incident = Incident(
            title=title,
            description=description,
            severity=severity,
            affected_services=affected_services or [],
            tags=tags or {},
            customer_impact=customer_impact,
            metadata=metadata,
        )
        
        # Add creation timeline entry
        incident.timeline.append(TimelineEntry(
            entry_type=TimelineEntryType.STATUS_CHANGE,
            message=f"Incident created with severity {severity.value}",
        ))
        
        await self._incident_store.save(incident)
        
        logger.info(f"Incident created: {incident.id} - {title}")
        
        # Notify on-call responders
        await self._notify_on_call(incident, "New incident created")
        
        # Notify listeners
        await self._notify_listeners("created", incident)
        
        return incident
    
    async def get(self, incident_id: str) -> Optional[Incident]:
        """Get incident."""
        return await self._incident_store.get(incident_id)
    
    async def update_status(
        self,
        incident_id: str,
        status: Union[str, IncidentStatus],
        message: str = "",
        author: str = "",
    ) -> Optional[Incident]:
        """Update incident status."""
        if isinstance(status, str):
            status = IncidentStatus(status)
        
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        old_status = incident.status
        incident.status = status
        
        if status == IncidentStatus.ACKNOWLEDGED and not incident.acknowledged_at:
            incident.acknowledged_at = datetime.utcnow()
        elif status == IncidentStatus.RESOLVED and not incident.resolved_at:
            incident.resolved_at = datetime.utcnow()
        elif status == IncidentStatus.CLOSED and not incident.closed_at:
            incident.closed_at = datetime.utcnow()
        
        incident.timeline.append(TimelineEntry(
            entry_type=TimelineEntryType.STATUS_CHANGE,
            message=message or f"Status changed from {old_status.value} to {status.value}",
            author=author,
        ))
        
        await self._incident_store.save(incident)
        
        logger.info(f"Incident {incident_id} status: {old_status.value} -> {status.value}")
        
        await self._notify_listeners("status_changed", incident)
        
        return incident
    
    async def add_timeline_entry(
        self,
        incident_id: str,
        message: str,
        entry_type: TimelineEntryType = TimelineEntryType.COMMENT,
        author: str = "",
        **metadata,
    ) -> Optional[TimelineEntry]:
        """Add timeline entry."""
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        entry = TimelineEntry(
            entry_type=entry_type,
            message=message,
            author=author,
            metadata=metadata,
        )
        
        incident.timeline.append(entry)
        
        await self._incident_store.save(incident)
        
        return entry
    
    async def assign(
        self,
        incident_id: str,
        assignee: str,
        author: str = "",
    ) -> Optional[Incident]:
        """Assign incident."""
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        old_assignee = incident.assignee
        incident.assignee = assignee
        
        incident.timeline.append(TimelineEntry(
            entry_type=TimelineEntryType.ASSIGNMENT,
            message=f"Assigned to {assignee}" + (f" from {old_assignee}" if old_assignee else ""),
            author=author,
        ))
        
        await self._incident_store.save(incident)
        
        return incident
    
    async def escalate(
        self,
        incident_id: str,
        level: Union[str, EscalationLevel],
        reason: str = "",
        author: str = "",
    ) -> Optional[Incident]:
        """Escalate incident."""
        if isinstance(level, str):
            level = EscalationLevel(level)
        
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        old_level = incident.escalation_level
        incident.escalation_level = level
        
        incident.timeline.append(TimelineEntry(
            entry_type=TimelineEntryType.ESCALATION,
            message=f"Escalated from {old_level.value} to {level.value}" + (f": {reason}" if reason else ""),
            author=author,
        ))
        
        await self._incident_store.save(incident)
        
        # Notify new level responders
        responders = await self._responder_store.get_on_call(level)
        await self._notifier.notify(incident, responders, f"Incident escalated to {level.value}")
        
        logger.info(f"Incident {incident_id} escalated to {level.value}")
        
        await self._notify_listeners("escalated", incident)
        
        return incident
    
    async def resolve(
        self,
        incident_id: str,
        resolution: str,
        root_cause: str = "",
        author: str = "",
    ) -> Optional[Incident]:
        """Resolve incident."""
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        incident.status = IncidentStatus.RESOLVED
        incident.resolved_at = datetime.utcnow()
        incident.resolution = resolution
        incident.root_cause = root_cause
        
        incident.timeline.append(TimelineEntry(
            entry_type=TimelineEntryType.STATUS_CHANGE,
            message=f"Incident resolved: {resolution}",
            author=author,
        ))
        
        await self._incident_store.save(incident)
        
        logger.info(f"Incident {incident_id} resolved")
        
        await self._notify_listeners("resolved", incident)
        
        return incident
    
    async def close(
        self,
        incident_id: str,
        author: str = "",
    ) -> Optional[Incident]:
        """Close incident."""
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        incident.status = IncidentStatus.CLOSED
        incident.closed_at = datetime.utcnow()
        
        incident.timeline.append(TimelineEntry(
            entry_type=TimelineEntryType.STATUS_CHANGE,
            message="Incident closed",
            author=author,
        ))
        
        await self._incident_store.save(incident)
        
        await self._notify_listeners("closed", incident)
        
        return incident
    
    async def list_incidents(
        self,
        status: Optional[List[IncidentStatus]] = None,
        severity: Optional[List[Severity]] = None,
        limit: int = 100,
    ) -> List[Incident]:
        """List incidents."""
        return await self._incident_store.list_incidents(status, severity, limit)
    
    async def get_open_incidents(self) -> List[Incident]:
        """Get open incidents."""
        return await self._incident_store.list_incidents(
            status=[
                IncidentStatus.OPEN,
                IncidentStatus.ACKNOWLEDGED,
                IncidentStatus.INVESTIGATING,
                IncidentStatus.IDENTIFIED,
                IncidentStatus.MONITORING,
            ]
        )
    
    async def add_responder(
        self,
        name: str,
        email: str,
        level: EscalationLevel = EscalationLevel.L1,
        on_call: bool = False,
        **kwargs,
    ) -> Responder:
        """Add responder."""
        responder = Responder(
            name=name,
            email=email,
            escalation_level=level,
            on_call=on_call,
            **kwargs,
        )
        
        await self._responder_store.save(responder)
        
        return responder
    
    async def set_on_call(
        self,
        responder_id: str,
        on_call: bool,
    ) -> Optional[Responder]:
        """Set responder on-call status."""
        responder = await self._responder_store.get(responder_id)
        
        if responder:
            responder.on_call = on_call
            await self._responder_store.save(responder)
        
        return responder
    
    async def get_stats(self, period_days: int = 30) -> IncidentStats:
        """Get incident statistics."""
        all_incidents = await self._incident_store.list_incidents(limit=10000)
        
        cutoff = datetime.utcnow() - timedelta(days=period_days)
        incidents = [i for i in all_incidents if i.created_at >= cutoff]
        
        stats = IncidentStats(
            total=len(incidents),
        )
        
        resolved_times = []
        ack_times = []
        
        for incident in incidents:
            # By status
            status = incident.status.value
            stats.by_status[status] = stats.by_status.get(status, 0) + 1
            
            if incident.status == IncidentStatus.RESOLVED or incident.status == IncidentStatus.CLOSED:
                stats.resolved += 1
            else:
                stats.open += 1
            
            # By severity
            severity = incident.severity.value
            stats.by_severity[severity] = stats.by_severity.get(severity, 0) + 1
            
            # Times
            if incident.time_to_resolve:
                resolved_times.append(incident.time_to_resolve.total_seconds())
            if incident.time_to_acknowledge:
                ack_times.append(incident.time_to_acknowledge.total_seconds())
        
        if resolved_times:
            stats.avg_time_to_resolve = sum(resolved_times) / len(resolved_times)
        if ack_times:
            stats.avg_time_to_acknowledge = sum(ack_times) / len(ack_times)
        
        return stats
    
    async def create_post_mortem(
        self,
        incident_id: str,
        summary: str = "",
        root_cause: str = "",
        lessons_learned: Optional[List[str]] = None,
        action_items: Optional[List[Dict[str, Any]]] = None,
        author: str = "",
    ) -> Optional[PostMortem]:
        """Create post-mortem for incident."""
        incident = await self._incident_store.get(incident_id)
        
        if not incident:
            return None
        
        post_mortem = PostMortem(
            incident_id=incident_id,
            title=f"Post-Mortem: {incident.title}",
            summary=summary or incident.description,
            timeline=incident.timeline.copy(),
            root_cause=root_cause or incident.root_cause,
            lessons_learned=lessons_learned or [],
            action_items=action_items or [],
            author=author,
        )
        
        return post_mortem
    
    def add_listener(self, listener: Callable) -> None:
        """Add event listener."""
        self._listeners.append(listener)
    
    async def _notify_on_call(self, incident: Incident, message: str) -> None:
        """Notify on-call responders."""
        responders = await self._responder_store.get_on_call(incident.escalation_level)
        
        if responders:
            await self._notifier.notify(incident, responders, message)
    
    async def _notify_listeners(self, event: str, incident: Incident) -> None:
        """Notify listeners."""
        for listener in self._listeners:
            try:
                if asyncio.iscoroutinefunction(listener):
                    await listener(event, incident)
                else:
                    listener(event, incident)
            except Exception as e:
                logger.error(f"Listener error: {e}")


# Factory functions
def create_incident_manager() -> IncidentManager:
    """Create incident manager."""
    return IncidentManager()


def create_incident(
    title: str,
    severity: Severity = Severity.MEDIUM,
    **kwargs,
) -> Incident:
    """Create incident."""
    return Incident(title=title, severity=severity, **kwargs)


def create_responder(
    name: str,
    email: str,
    **kwargs,
) -> Responder:
    """Create responder."""
    return Responder(name=name, email=email, **kwargs)


__all__ = [
    # Exceptions
    "IncidentError",
    # Enums
    "Severity",
    "IncidentStatus",
    "EscalationLevel",
    "TimelineEntryType",
    # Data classes
    "TimelineEntry",
    "Responder",
    "EscalationPolicy",
    "Incident",
    "IncidentStats",
    "PostMortem",
    # Stores
    "IncidentStore",
    "InMemoryIncidentStore",
    "ResponderStore",
    "InMemoryResponderStore",
    # Notifier
    "Notifier",
    "LogNotifier",
    # Manager
    "IncidentManager",
    # Factory functions
    "create_incident_manager",
    "create_incident",
    "create_responder",
]
